# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DQN Builder."""
from typing import Callable, Iterator, List, Optional, Sequence

from acme import adders
from acme import core
from acme import specs
from acme.adders import reverb as adders_reverb
from acme.agents.jax import actors
from acme.agents.jax import builders
from acme.agents.jax.dqn import actor as dqn_actor
from acme.agents.jax.dqn import config as dqn_config
from acme.agents.jax.dqn import learning_lib
from acme.datasets import reverb as datasets
from acme.jax import networks as networks_lib
from acme.jax import variable_utils
from acme.utils import counting
from acme.utils import loggers
import optax
import reverb
from reverb import rate_limiters


class DQNBuilder(builders.ActorLearnerBuilder):
  """DQN Builder."""

  def __init__(
      self,
      config: dqn_config.DQNConfig,
      loss_fn: learning_lib.LossFn,
      logger_fn: Callable[[], loggers.Logger] = lambda: None,
      actor_backend: Optional[str] = 'cpu'
  ):
    """Creates DQN learner and the behavior policies.

    Args:
      config: DQN config.
      loss_fn: A loss function.
      logger_fn: a logger factory for the learner.
      actor_backend: Which backend to use when jitting the policy.
    """
    self._config = config
    self._loss_fn = loss_fn
    self._logger_fn = logger_fn
    self._actor_backend = actor_backend

  def make_learner(
      self,
      random_key: networks_lib.PRNGKey,
      networks: networks_lib.FeedForwardNetwork,
      dataset: Iterator[reverb.ReplaySample],
      replay_client: Optional[reverb.Client] = None,
      counter: Optional[counting.Counter] = None,
  ) -> core.Learner:
    return learning_lib.SGDLearner(
        network=networks,
        random_key=random_key,
        optimizer=optax.adam(self._config.learning_rate,
                             eps=self._config.adam_eps),
        target_update_period=self._config.target_update_period,
        data_iterator=dataset,
        loss_fn=self._loss_fn,
        replay_client=replay_client,
        replay_table_name=self._config.replay_table_name,
        counter=counter,
        num_sgd_steps_per_step=self._config.num_sgd_steps_per_step,
        logger=self._logger_fn())

  def make_actor(
      self,
      random_key: networks_lib.PRNGKey,
      policy_network: dqn_actor.EpsilonPolicy,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None,
  ) -> core.Actor:
    assert variable_source is not None
    # Inference happens on CPU, so it's better to move variables there too.
    variable_client = variable_utils.VariableClient(variable_source, '',
                                                    device='cpu')
    epsilon = self._config.epsilon
    epsilons = epsilon if epsilon is Sequence else (epsilon,)
    actor_core = dqn_actor.alternating_epsilons_actor_core(
        policy_network, epsilons=epsilons)
    return actors.GenericActor(actor=actor_core,
                               random_key=random_key,
                               variable_client=variable_client,
                               adder=adder,
                               backend=self._actor_backend)

  def make_replay_tables(
      self, environment_spec: specs.EnvironmentSpec) -> List[reverb.Table]:
    """Creates reverb tables for the algorithm."""
    samples_per_insert_tolerance = (
        self._config.samples_per_insert_tolerance_rate *
        self._config.samples_per_insert)
    error_buffer = self._config.min_replay_size * samples_per_insert_tolerance
    limiter = rate_limiters.SampleToInsertRatio(
        min_size_to_sample=self._config.min_replay_size,
        samples_per_insert=self._config.samples_per_insert,
        error_buffer=error_buffer)
    return [reverb.Table(
        name=self._config.replay_table_name,
        sampler=reverb.selectors.Prioritized(self._config.priority_exponent),
        remover=reverb.selectors.Fifo(),
        max_size=self._config.max_replay_size,
        rate_limiter=limiter,
        signature=adders_reverb.NStepTransitionAdder.signature(
            environment_spec))]

  def make_dataset_iterator(
      self, replay_client: reverb.Client) -> Iterator[reverb.ReplaySample]:
    """Creates a dataset iterator to use for learning."""
    dataset = datasets.make_reverb_dataset(
        table=self._config.replay_table_name,
        server_address=replay_client.server_address,
        batch_size=(
            self._config.batch_size * self._config.num_sgd_steps_per_step),
        prefetch_size=self._config.prefetch_size)
    return dataset.as_numpy_iterator()

  def make_adder(self, replay_client: reverb.Client) -> adders.Adder:
    """Creates an adder which handles observations."""
    return adders_reverb.NStepTransitionAdder(
        priority_fns={self._config.replay_table_name: None},
        client=replay_client,
        n_step=self._config.n_step,
        discount=self._config.discount)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for core.py."""

from typing import List

from absl.testing import absltest
from acme import core
from acme import types


class StepCountingLearner(core.Learner):
  """A learner which counts `num_steps` and then raises `StopIteration`."""

  def __init__(self, num_steps: int):
    self.step_count = 0
    self.num_steps = num_steps

  def step(self):
    self.step_count += 1
    if self.step_count >= self.num_steps:
      raise StopIteration()

  def get_variables(self, unused: List[str]) -> List[types.NestedArray]:
    del unused
    return []


class CoreTest(absltest.TestCase):

  def test_learner_run_with_limit(self):
    learner = StepCountingLearner(100)
    learner.run(7)
    self.assertEqual(learner.step_count, 7)

  def test_learner_run_no_limit(self):
    learner = StepCountingLearner(100)
    with self.assertRaises(StopIteration):
      learner.run()
    self.assertEqual(learner.step_count, 100)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Recurrent CRR learner implementation."""

import operator
import time
from typing import Dict, List, Optional

from acme import core
from acme.tf import losses
from acme.tf import networks
from acme.tf import savers as tf2_savers
from acme.tf import utils as tf2_utils
from acme.utils import counting
from acme.utils import loggers
import numpy as np
import reverb
import sonnet as snt
import tensorflow as tf
import tree


class RCRRLearner(core.Learner):
  """Recurrent CRR learner.

  This is the learning component of a RCRR agent. It takes a dataset as
  input and implements update functionality to learn from this dataset.
  """

  def __init__(self,
               policy_network: snt.RNNCore,
               critic_network: networks.CriticDeepRNN,
               target_policy_network: snt.RNNCore,
               target_critic_network: networks.CriticDeepRNN,
               dataset: tf.data.Dataset,
               accelerator_strategy: Optional[tf.distribute.Strategy] = None,
               behavior_network: Optional[snt.Module] = None,
               cwp_network: Optional[snt.Module] = None,
               policy_optimizer: Optional[snt.Optimizer] = None,
               critic_optimizer: Optional[snt.Optimizer] = None,
               discount: float = 0.99,
               target_update_period: int = 100,
               num_action_samples_td_learning: int = 1,
               num_action_samples_policy_weight: int = 4,
               baseline_reduce_function: str = 'mean',
               clipping: bool = True,
               policy_improvement_modes: str = 'exp',
               ratio_upper_bound: float = 20.,
               beta: float = 1.0,
               counter: Optional[counting.Counter] = None,
               logger: Optional[loggers.Logger] = None,
               checkpoint: bool = False):
    """Initializes the learner.

    Args:
      policy_network: the online (optimized) policy.
      critic_network: the online critic.
      target_policy_network: the target policy (which lags behind the online
        policy).
      target_critic_network: the target critic.
      dataset: dataset to learn from, whether fixed or from a replay buffer
        (see `acme.datasets.reverb.make_dataset` documentation).
      accelerator_strategy: the strategy used to distribute computation,
        whether on a single, or multiple, GPU or TPU; as supported by
        tf.distribute.
      behavior_network: The network to snapshot under `policy` name. If None,
        snapshots `policy_network` instead.
      cwp_network: CWP network to snapshot: samples actions
        from the policy and weighs them with the critic, then returns the action
        by sampling from the softmax distribution using critic values as logits.
        Used only for snapshotting, not training.
      policy_optimizer: the optimizer to be applied to the policy loss.
      critic_optimizer: the optimizer to be applied to the distributional
        Bellman loss.
      discount: discount to use for TD updates.
      target_update_period: number of learner steps to perform before updating
        the target networks.
      num_action_samples_td_learning: number of action samples to use to
        estimate expected value of the critic loss w.r.t. stochastic policy.
      num_action_samples_policy_weight: number of action samples to use to
        estimate the advantage function for the CRR weighting of the policy
        loss.
      baseline_reduce_function: one of 'mean', 'max', 'min'. Way of aggregating
        values from `num_action_samples` estimates of the value function.
      clipping: whether to clip gradients by global norm.
      policy_improvement_modes: one of 'exp', 'binary', 'all'. CRR mode which
        determines how the advantage function is processed before being
        multiplied by the policy loss.
      ratio_upper_bound: if policy_improvement_modes is 'exp', determines
        the upper bound of the weight (i.e. the weight is
          min(exp(advantage / beta), upper_bound)
        ).
      beta: if policy_improvement_modes is 'exp', determines the beta (see
        above).
      counter: counter object used to keep track of steps.
      logger: logger object to be used by learner.
      checkpoint: boolean indicating whether to checkpoint the learner.
    """

    if accelerator_strategy is None:
      accelerator_strategy = snt.distribute.Replicator()
    self._accelerator_strategy = accelerator_strategy
    self._policy_improvement_modes = policy_improvement_modes
    self._ratio_upper_bound = ratio_upper_bound
    self._num_action_samples_td_learning = num_action_samples_td_learning
    self._num_action_samples_policy_weight = num_action_samples_policy_weight
    self._baseline_reduce_function = baseline_reduce_function
    self._beta = beta

    # When running on TPUs we have to know the amount of memory required (and
    # thus the sequence length) at the graph compilation stage. At the moment,
    # the only way to get it is to sample from the dataset, since the dataset
    # does not have any metadata, see b/160672927 to track this upcoming
    # feature.
    sample = next(dataset.as_numpy_iterator())
    self._sequence_length = sample.action.shape[1]

    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.TerminalLogger('learner', time_delta=1.)
    self._discount = discount
    self._clipping = clipping

    self._target_update_period = target_update_period

    with self._accelerator_strategy.scope():
      # Necessary to track when to update target networks.
      self._num_steps = tf.Variable(0, dtype=tf.int32)

      # (Maybe) distributing the dataset across multiple accelerators.
      distributed_dataset = self._accelerator_strategy.experimental_distribute_dataset(
          dataset)
      self._iterator = iter(distributed_dataset)

      # Create the optimizers.
      self._critic_optimizer = critic_optimizer or snt.optimizers.Adam(1e-4)
      self._policy_optimizer = policy_optimizer or snt.optimizers.Adam(1e-4)

    # Store online and target networks.
    self._policy_network = policy_network
    self._critic_network = critic_network
    self._target_policy_network = target_policy_network
    self._target_critic_network = target_critic_network

    # Expose the variables.
    self._variables = {
        'critic': self._target_critic_network.variables,
        'policy': self._target_policy_network.variables,
    }

    # Create a checkpointer object.
    self._checkpointer = None
    self._snapshotter = None
    if checkpoint:
      self._checkpointer = tf2_savers.Checkpointer(
          objects_to_save={
              'counter': self._counter,
              'policy': self._policy_network,
              'critic': self._critic_network,
              'target_policy': self._target_policy_network,
              'target_critic': self._target_critic_network,
              'policy_optimizer': self._policy_optimizer,
              'critic_optimizer': self._critic_optimizer,
              'num_steps': self._num_steps,
          },
          time_delta_minutes=30.)

      raw_policy = snt.DeepRNN(
          [policy_network, networks.StochasticSamplingHead()])
      critic_mean = networks.CriticDeepRNN(
          [critic_network, networks.StochasticMeanHead()])
      objects_to_save = {
          'raw_policy': raw_policy,
          'critic': critic_mean,
      }
      if behavior_network is not None:
        objects_to_save['policy'] = behavior_network
      if cwp_network is not None:
        objects_to_save['cwp_policy'] = cwp_network
      self._snapshotter = tf2_savers.Snapshotter(
          objects_to_save=objects_to_save, time_delta_minutes=30)
    # Timestamp to keep track of the wall time.
    self._walltime_timestamp = time.time()

  def _step(self, sample: reverb.ReplaySample) -> Dict[str, tf.Tensor]:
    # Transpose batch and sequence axes, i.e. [B, T, ...] to [T, B, ...].
    sample = tf2_utils.batch_to_sequence(sample)
    observations = sample.observation
    actions = sample.action
    rewards = sample.reward
    discounts = sample.discount

    dtype = rewards.dtype

    # Cast the additional discount to match the environment discount dtype.
    discount = tf.cast(self._discount, dtype=discounts.dtype)

    # Loss cumulants across time. These cannot be python mutable objects.
    critic_loss = 0.
    policy_loss = 0.

    # Each transition induces a policy loss, which we then weight using
    # the `policy_loss_coef_t`; shape [B], see https://arxiv.org/abs/2006.15134.
    # `policy_loss_coef` is a scalar average of these coefficients across
    # the batch and sequence length dimensions.
    policy_loss_coef = 0.

    per_device_batch_size = actions.shape[1]

    # Initialize recurrent states.
    critic_state = self._critic_network.initial_state(per_device_batch_size)
    target_critic_state = critic_state
    policy_state = self._policy_network.initial_state(per_device_batch_size)
    target_policy_state = policy_state

    with tf.GradientTape(persistent=True) as tape:
      for t in range(1, self._sequence_length):
        o_tm1 = tree.map_structure(operator.itemgetter(t - 1), observations)
        a_tm1 = tree.map_structure(operator.itemgetter(t - 1), actions)
        r_t = tree.map_structure(operator.itemgetter(t - 1), rewards)
        d_t = tree.map_structure(operator.itemgetter(t - 1), discounts)
        o_t = tree.map_structure(operator.itemgetter(t), observations)

        if t != 1:
          # By only updating the target critic state here we are forcing
          # the target critic to ignore observations[0]. Otherwise, the
          # target_critic will be unrolled for one more timestep than critic.
          # The smaller the sequence length, the more problematic this is: if
          # you use RNN on sequences of length 2, you would expect the code to
          # never use recurrent connections. But if you don't skip updating the
          # target_critic_state on observation[0] here, it won't be the case.
          _, target_critic_state = self._target_critic_network(
              o_tm1, a_tm1, target_critic_state)

        # ========================= Critic learning ============================
        q_tm1, next_critic_state = self._critic_network(o_tm1, a_tm1,
                                                        critic_state)
        target_action_distribution, target_policy_state = self._target_policy_network(
            o_t, target_policy_state)

        sampled_actions_t = target_action_distribution.sample(
            self._num_action_samples_td_learning)
        # [N, B, ...]
        tiled_o_t = tf2_utils.tile_nested(
            o_t, self._num_action_samples_td_learning)
        tiled_target_critic_state = tf2_utils.tile_nested(
            target_critic_state, self._num_action_samples_td_learning)

        # Compute the target critic's Q-value of the sampled actions.
        sampled_q_t, _ = snt.BatchApply(self._target_critic_network)(
            tiled_o_t, sampled_actions_t, tiled_target_critic_state)

        # Compute average logits by first reshaping them to [N, B, A] and then
        # normalizing them across atoms.
        new_shape = [self._num_action_samples_td_learning, r_t.shape[0], -1]
        sampled_logits = tf.reshape(sampled_q_t.logits, new_shape)
        sampled_logprobs = tf.math.log_softmax(sampled_logits, axis=-1)
        averaged_logits = tf.reduce_logsumexp(sampled_logprobs, axis=0)

        # Construct the expected distributional value for bootstrapping.
        q_t = networks.DiscreteValuedDistribution(
            values=sampled_q_t.values, logits=averaged_logits)
        critic_loss_t = losses.categorical(q_tm1, r_t, discount * d_t, q_t)
        critic_loss_t = tf.reduce_mean(critic_loss_t)

        # ========================= Actor learning =============================
        action_distribution_tm1, policy_state = self._policy_network(
            o_tm1, policy_state)
        q_tm1_mean = q_tm1.mean()

        # Compute the estimate of the value function based on
        # self._num_action_samples_policy_weight samples from the policy.
        tiled_o_tm1 = tf2_utils.tile_nested(
            o_tm1, self._num_action_samples_policy_weight)
        tiled_critic_state = tf2_utils.tile_nested(
            critic_state, self._num_action_samples_policy_weight)
        action_tm1 = action_distribution_tm1.sample(
            self._num_action_samples_policy_weight)
        tiled_z_tm1, _ = snt.BatchApply(self._critic_network)(
            tiled_o_tm1, action_tm1, tiled_critic_state)
        tiled_v_tm1 = tf.reshape(tiled_z_tm1.mean(),
                                 [self._num_action_samples_policy_weight, -1])

        # Use mean, min, or max to aggregate Q(s, a_i), a_i ~ pi(s) into the
        # final estimate of the value function.
        if self._baseline_reduce_function == 'mean':
          v_tm1_estimate = tf.reduce_mean(tiled_v_tm1, axis=0)
        elif self._baseline_reduce_function == 'max':
          v_tm1_estimate = tf.reduce_max(tiled_v_tm1, axis=0)
        elif self._baseline_reduce_function == 'min':
          v_tm1_estimate = tf.reduce_min(tiled_v_tm1, axis=0)

        # Assert that action_distribution_tm1 is a batch of multivariate
        # distributions (in contrast to e.g. a [batch, action_size] collection
        # of 1d distributions).
        assert len(action_distribution_tm1.batch_shape) == 1
        policy_loss_batch = -action_distribution_tm1.log_prob(a_tm1)

        advantage = q_tm1_mean - v_tm1_estimate
        if self._policy_improvement_modes == 'exp':
          policy_loss_coef_t = tf.math.minimum(
              tf.math.exp(advantage / self._beta), self._ratio_upper_bound)
        elif self._policy_improvement_modes == 'binary':
          policy_loss_coef_t = tf.cast(advantage > 0, dtype=dtype)
        elif self._policy_improvement_modes == 'all':
          # Regress against all actions (effectively pure BC).
          policy_loss_coef_t = 1.
        policy_loss_coef_t = tf.stop_gradient(policy_loss_coef_t)

        policy_loss_batch *= policy_loss_coef_t
        policy_loss_t = tf.reduce_mean(policy_loss_batch)

        critic_state = next_critic_state

        critic_loss += critic_loss_t
        policy_loss += policy_loss_t
        policy_loss_coef += tf.reduce_mean(policy_loss_coef_t)  # For logging.

      # Divide by sequence length to get mean losses.
      critic_loss /= tf.cast(self._sequence_length, dtype=dtype)
      policy_loss /= tf.cast(self._sequence_length, dtype=dtype)
      policy_loss_coef /= tf.cast(self._sequence_length, dtype=dtype)

    # Compute gradients.
    critic_gradients = tape.gradient(critic_loss,
                                     self._critic_network.trainable_variables)
    policy_gradients = tape.gradient(policy_loss,
                                     self._policy_network.trainable_variables)

    # Delete the tape manually because of the persistent=True flag.
    del tape

    # Sync gradients across GPUs or TPUs.
    ctx = tf.distribute.get_replica_context()
    critic_gradients = ctx.all_reduce('mean', critic_gradients)
    policy_gradients = ctx.all_reduce('mean', policy_gradients)

    # Maybe clip gradients.
    if self._clipping:
      policy_gradients = tf.clip_by_global_norm(policy_gradients, 40.)[0]
      critic_gradients = tf.clip_by_global_norm(critic_gradients, 40.)[0]

    # Apply gradients.
    self._critic_optimizer.apply(critic_gradients,
                                 self._critic_network.trainable_variables)
    self._policy_optimizer.apply(policy_gradients,
                                 self._policy_network.trainable_variables)

    source_variables = (
        self._critic_network.variables + self._policy_network.variables)
    target_variables = (
        self._target_critic_network.variables +
        self._target_policy_network.variables)

    # Make online -> target network update ops.
    if tf.math.mod(self._num_steps, self._target_update_period) == 0:
      for src, dest in zip(source_variables, target_variables):
        dest.assign(src)
    self._num_steps.assign_add(1)

    return {
        'critic_loss': critic_loss,
        'policy_loss': policy_loss,
        'policy_loss_coef': policy_loss_coef,
    }

  @tf.function
  def _replicated_step(self) -> Dict[str, tf.Tensor]:
    sample = next(self._iterator)
    fetches = self._accelerator_strategy.run(self._step, args=(sample,))
    mean = tf.distribute.ReduceOp.MEAN
    return {
        k: self._accelerator_strategy.reduce(mean, fetches[k], axis=None)
        for k in fetches
    }

  def step(self):
    # Run the learning step.
    with self._accelerator_strategy.scope():
      fetches = self._replicated_step()

    # Update our counts and record it.
    new_timestamp = time.time()
    time_passed = new_timestamp - self._walltime_timestamp
    self._walltime_timestamp = new_timestamp
    counts = self._counter.increment(steps=1, wall_time=time_passed)
    fetches.update(counts)

    # Checkpoint and attempt to write the logs.
    if self._checkpointer is not None:
      self._checkpointer.save()
      self._snapshotter.save()
    self._logger.write(fetches)

  def get_variables(self, names: List[str]) -> List[List[np.ndarray]]:
    return [tf2_utils.to_numpy(self._variables[name]) for name in names]



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Haiku modules that output tfd.Distributions."""

from typing import Any, List, Optional, Union

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import tensorflow_probability
hk_init = hk.initializers
tfp = tensorflow_probability.substrates.jax
tfd = tfp.distributions

_MIN_SCALE = 1e-4
Initializer = hk.initializers.Initializer


class CategoricalHead(hk.Module):
  """Module that produces a categorical distribution with the given number of values."""

  def __init__(
      self,
      num_values: Union[int, List[int]],
      dtype: Optional[Any] = jnp.int32,
      w_init: Optional[Initializer] = None,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self._dtype = dtype
    self._logit_shape = num_values
    self._linear = hk.Linear(np.prod(num_values), w_init=w_init)

  def __call__(self, inputs: jnp.ndarray) -> tfd.Distribution:
    logits = self._linear(inputs)
    if not isinstance(self._logit_shape, int):
      logits = hk.Reshape(self._logit_shape)(logits)
    return tfd.Categorical(logits=logits, dtype=self._dtype)


class GaussianMixture(hk.Module):
  """Module that outputs a Gaussian Mixture Distribution."""

  def __init__(self,
               num_dimensions: int,
               num_components: int,
               multivariate: bool,
               init_scale: Optional[float] = None,
               append_singleton_event_dim: bool = False,
               reinterpreted_batch_ndims: Optional[int] = None,
               name: str = 'GaussianMixture'):
    """Initialization.

    Args:
      num_dimensions: dimensionality of the output distribution
      num_components: number of mixture components.
      multivariate: whether the resulting distribution is multivariate or not.
      init_scale: the initial scale for the Gaussian mixture components.
      append_singleton_event_dim: (univariate only) Whether to add an extra
        singleton dimension to the event shape.
      reinterpreted_batch_ndims: (univariate only) Number of batch dimensions to
        reinterpret as event dimensions.
      name: name of the module passed to snt.Module parent class.
    """
    super().__init__(name=name)

    self._num_dimensions = num_dimensions
    self._num_components = num_components
    self._multivariate = multivariate
    self._append_singleton_event_dim = append_singleton_event_dim
    self._reinterpreted_batch_ndims = reinterpreted_batch_ndims

    if init_scale is not None:
      self._scale_factor = init_scale / jax.nn.softplus(0.)
    else:
      self._scale_factor = 1.0  # Corresponds to init_scale = softplus(0).

  def __call__(self,
               inputs: jnp.ndarray,
               low_noise_policy: bool = False) -> tfd.Distribution:
    """Run the networks through inputs.

    Args:
      inputs: hidden activations of the policy network body.
      low_noise_policy: whether to set vanishingly small scales for each
        component. If this flag is set to True, the policy is effectively run
        without Gaussian noise.

    Returns:
      Mixture Gaussian distribution.
    """

    # Define the weight initializer.
    w_init = hk.initializers.VarianceScaling(scale=1e-5)

    # Create a layer that outputs the unnormalized log-weights.
    if self._multivariate:
      logits_size = self._num_components
    else:
      logits_size = self._num_dimensions * self._num_components
    logit_layer = hk.Linear(logits_size, w_init=w_init)

    # Create two layers that outputs a location and a scale, respectively, for
    # each dimension and each component.
    loc_layer = hk.Linear(
        self._num_dimensions * self._num_components, w_init=w_init)
    scale_layer = hk.Linear(
        self._num_dimensions * self._num_components, w_init=w_init)

    # Compute logits, locs, and scales if necessary.
    logits = logit_layer(inputs)
    locs = loc_layer(inputs)

    # When a low_noise_policy is requested, set the scales to its minimum value.
    if low_noise_policy:
      scales = jnp.full(locs.shape, _MIN_SCALE)
    else:
      scales = scale_layer(inputs)
      scales = self._scale_factor * jax.nn.softplus(scales) + _MIN_SCALE

    if self._multivariate:
      components_class = tfd.MultivariateNormalDiag
      shape = [-1, self._num_components, self._num_dimensions]  # [B, C, D]
      # In this case, no need to reshape logits as they are in the correct shape
      # already, namely [batch_size, num_components].
    else:
      components_class = tfd.Normal
      shape = [-1, self._num_dimensions, self._num_components]  # [B, D, C]
      if self._append_singleton_event_dim:
        shape.insert(2, 1)  # [B, D, 1, C]
      logits = logits.reshape(shape)

    # Reshape the mixture's location and scale parameters appropriately.
    locs = locs.reshape(shape)
    scales = scales.reshape(shape)

    # Create the mixture distribution.
    distribution = tfd.MixtureSameFamily(
        mixture_distribution=tfd.Categorical(logits=logits),
        components_distribution=components_class(loc=locs, scale=scales))

    if not self._multivariate:
      distribution = tfd.Independent(
          distribution,
          reinterpreted_batch_ndims=self._reinterpreted_batch_ndims)

    return distribution


class TanhTransformedDistribution(tfd.TransformedDistribution):
  """Distribution followed by tanh."""

  def __init__(self, distribution, threshold=.999, validate_args=False):
    """Initialize the distribution.

    Args:
      distribution: The distribution to transform.
      threshold: Clipping value of the action when computing the logprob.
      validate_args: Passed to super class.
    """
    super().__init__(
        distribution=distribution,
        bijector=tfp.bijectors.Tanh(),
        validate_args=validate_args)
    # Computes the log of the average probability distribution outside the
    # clipping range, i.e. on the interval [-inf, -atanh(threshold)] for
    # log_prob_left and [atanh(threshold), inf] for log_prob_right.
    self._threshold = threshold
    inverse_threshold = self.bijector.inverse(threshold)
    # average(pdf) = p/epsilon
    # So log(average(pdf)) = log(p) - log(epsilon)
    log_epsilon = jnp.log(1. - threshold)
    # Those 2 values are differentiable w.r.t. model parameters, such that the
    # gradient is defined everywhere.
    self._log_prob_left = self.distribution.log_cdf(
        -inverse_threshold) - log_epsilon
    self._log_prob_right = self.distribution.log_survival_function(
        inverse_threshold) - log_epsilon

  def log_prob(self, event):
    # Without this clip there would be NaNs in the inner tf.where and that
    # causes issues for some reasons.
    event = jnp.clip(event, -self._threshold, self._threshold)
    # The inverse image of {threshold} is the interval [atanh(threshold), inf]
    # which has a probability of "log_prob_right" under the given distribution.
    return jnp.where(
        event <= -self._threshold, self._log_prob_left,
        jnp.where(event >= self._threshold, self._log_prob_right,
                  super().log_prob(event)))

  def mode(self):
    return self.bijector.forward(self.distribution.mode())

  @classmethod
  def _parameter_properties(cls, dtype: Optional[Any], num_classes=None):
    td_properties = super()._parameter_properties(dtype,
                                                  num_classes=num_classes)
    del td_properties['bijector']
    return td_properties


class NormalTanhDistribution(hk.Module):
  """Module that produces a TanhTransformedDistribution distribution."""

  def __init__(self,
               num_dimensions: int,
               min_scale: float = 1e-3,
               w_init: hk_init.Initializer = hk_init.VarianceScaling(
                   1.0, 'fan_in', 'uniform'),
               b_init: hk_init.Initializer = hk_init.Constant(0.)):
    """Initialization.

    Args:
      num_dimensions: Number of dimensions of a distribution.
      min_scale: Minimum standard deviation.
      w_init: Initialization for linear layer weights.
      b_init: Initialization for linear layer biases.
    """
    super().__init__(name='Normal')
    self._min_scale = min_scale
    self._loc_layer = hk.Linear(num_dimensions, w_init=w_init, b_init=b_init)
    self._scale_layer = hk.Linear(num_dimensions, w_init=w_init, b_init=b_init)

  def __call__(self, inputs: jnp.ndarray) -> tfd.Distribution:
    loc = self._loc_layer(inputs)
    scale = self._scale_layer(inputs)
    scale = jax.nn.softplus(scale) + self._min_scale
    distribution = tfd.Normal(loc=loc, scale=scale)
    return tfd.Independent(
        TanhTransformedDistribution(distribution), reinterpreted_batch_ndims=1)


class MultivariateNormalDiagHead(hk.Module):
  """Module that produces a tfd.MultivariateNormalDiag distribution."""

  def __init__(self,
               num_dimensions: int,
               min_scale: float = 1e-6,
               w_init: hk_init.Initializer = hk_init.VarianceScaling(1e-4),
               b_init: hk_init.Initializer = hk_init.Constant(0.)):
    """Initialization.

    Args:
      num_dimensions: Number of dimensions of MVN distribution.
      min_scale: Minimum standard deviation.
      w_init: Initialization for linear layer weights.
      b_init: Initialization for linear layer biases.
    """
    super().__init__(name='MultivariateNormalDiagHead')
    self._min_scale = min_scale
    self._loc_layer = hk.Linear(num_dimensions, w_init=w_init, b_init=b_init)
    self._scale_layer = hk.Linear(num_dimensions, w_init=w_init, b_init=b_init)

  def __call__(self, inputs: jnp.ndarray) -> tfd.Distribution:
    loc = self._loc_layer(inputs)
    scale = self._scale_layer(inputs)
    scale = jax.nn.softplus(scale) + self._min_scale
    return tfd.MultivariateNormalDiag(loc=loc, scale_diag=scale)


class CategoricalValueHead(hk.Module):
  """Network head that produces a categorical distribution and value."""

  def __init__(
      self,
      num_values: int,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self._logit_layer = hk.Linear(num_values)
    self._value_layer = hk.Linear(1)

  def __call__(self, inputs: jnp.ndarray):
    logits = self._logit_layer(inputs)
    value = jnp.squeeze(self._value_layer(inputs), axis=-1)
    return (tfd.Categorical(logits=logits), value)


class DiscreteValued(hk.Module):
  """C51-style head.

  For each action, it produces the logits for a discrete distribution over
  atoms. Therefore, the returned logits represents several distributions, one
  for each action.
  """

  def __init__(
      self,
      num_actions: int,
      head_units: int = 512,
      num_atoms: int = 51,
      v_min: float = -1.0,
      v_max: float = 1.0,
  ):
    super().__init__('DiscreteValued')
    self._num_actions = num_actions
    self._num_atoms = num_atoms
    self._atoms = jnp.linspace(v_min, v_max, self._num_atoms)
    self._network = hk.nets.MLP([head_units, num_actions * num_atoms])

  def __call__(self, inputs: jnp.ndarray):
    q_logits = self._network(inputs)
    q_logits = jnp.reshape(q_logits, (-1, self._num_actions, self._num_atoms))
    q_dist = jax.nn.softmax(q_logits)
    q_values = jnp.sum(q_dist * self._atoms, axis=2)
    q_values = jax.lax.stop_gradient(q_values)
    return q_values, q_logits, self._atoms



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BC learner implementation."""

import time
from typing import Dict, Iterator, List, NamedTuple, Optional, Tuple

import acme
from acme import types
from acme.agents.jax.bc import losses
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.utils import counting
from acme.utils import loggers
import jax
import jax.numpy as jnp
import optax


class TrainingState(NamedTuple):
  """Contains training state for the learner."""
  optimizer_state: optax.OptState
  policy_params: networks_lib.Params
  key: networks_lib.PRNGKey
  steps: int


class BCLearner(acme.Learner):
  """BC learner.

  This is the learning component of a BC agent. It takes a Transitions iterator
  as input and implements update functionality to learn from this iterator.
  """

  _state: TrainingState

  def __init__(self,
               network: networks_lib.FeedForwardNetwork,
               random_key: networks_lib.PRNGKey,
               loss_fn: losses.Loss,
               optimizer: optax.GradientTransformation,
               demonstrations: Iterator[types.Transition],
               num_sgd_steps_per_step: int,
               logger: Optional[loggers.Logger] = None,
               counter: Optional[counting.Counter] = None):
    """Behavior Cloning Learner.

    Args:
      network: Networks with signature for apply:
        (params, obs, is_training, key) -> jnp.ndarray
        and for init:
        (rng, is_training) -> params
      random_key: RNG key.
      loss_fn: BC loss to use.
      optimizer: Optax optimizer.
      demonstrations: Demonstrations iterator.
      num_sgd_steps_per_step: Number of gradient updates per step.
      logger: Logger
      counter: Counter
    """
    def sgd_step(
        state: TrainingState,
        transitions: types.Transition,
    ) -> Tuple[TrainingState, Dict[str, jnp.ndarray]]:

      loss_and_grad = jax.value_and_grad(loss_fn, argnums=1)

      # Compute losses and their gradients.
      key, key_input = jax.random.split(state.key)
      loss_value, gradients = loss_and_grad(network.apply, state.policy_params,
                                            key_input, transitions)

      policy_update, optimizer_state = optimizer.update(
          gradients, state.optimizer_state, state.policy_params)
      policy_params = optax.apply_updates(state.policy_params, policy_update)

      new_state = TrainingState(
          optimizer_state=optimizer_state,
          policy_params=policy_params,
          key=key,
          steps=state.steps + 1,
      )
      metrics = {
          'loss': loss_value,
          'gradient_norm': optax.global_norm(gradients)
      }

      return new_state, metrics

    # General learner book-keeping and loggers.
    self._counter = counter or counting.Counter(prefix='learner')
    self._logger = logger or loggers.make_default_logger(
        'learner', asynchronous=True, serialize_fn=utils.fetch_devicearray)

    # Iterator on demonstration transitions.
    self._demonstrations = demonstrations

    # Split the input batch to `num_sgd_steps_per_step` minibatches in order
    # to achieve better performance on accelerators.
    self._sgd_step = jax.jit(utils.process_multiple_batches(
        sgd_step, num_sgd_steps_per_step))

    random_key, init_key = jax.random.split(random_key)
    policy_params = network.init(init_key)
    optimizer_state = optimizer.init(policy_params)

    # Create initial state.
    self._state = TrainingState(
        optimizer_state=optimizer_state,
        policy_params=policy_params,
        key=random_key,
        steps=0,
    )

    self._timestamp = None

  def step(self):
    # Get a batch of Transitions.
    transitions = next(self._demonstrations)
    self._state, metrics = self._sgd_step(self._state, transitions)

    # Compute elapsed time.
    timestamp = time.time()
    elapsed_time = timestamp - self._timestamp if self._timestamp else 0
    self._timestamp = timestamp

    # Increment counts and record the current time
    counts = self._counter.increment(steps=1, walltime=elapsed_time)

    # Attempts to write the logs.
    self._logger.write({**metrics, **counts})

  def get_variables(self, names: List[str]) -> List[networks_lib.Params]:
    variables = {
        'policy': self._state.policy_params,
    }
    return [variables[name] for name in names]

  def save(self) -> TrainingState:
    return self._state

  def restore(self, state: TrainingState):
    self._state = state



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DQN learner implementation."""

import time
from typing import Dict, List, Optional, Union

import acme
from acme import types
from acme.adders import reverb as adders
from acme.tf import losses
from acme.tf import savers as tf2_savers
from acme.tf import utils as tf2_utils
from acme.utils import counting
from acme.utils import loggers
import numpy as np
import reverb
import sonnet as snt
import tensorflow as tf
import trfl


class DQNLearner(acme.Learner, tf2_savers.TFSaveable):
  """DQN learner.

  This is the learning component of a DQN agent. It takes a dataset as input
  and implements update functionality to learn from this dataset. Optionally
  it takes a replay client as well to allow for updating of priorities.
  """

  def __init__(
      self,
      network: snt.Module,
      target_network: snt.Module,
      discount: float,
      importance_sampling_exponent: float,
      learning_rate: float,
      target_update_period: int,
      dataset: tf.data.Dataset,
      huber_loss_parameter: float = 1.,
      replay_client: Optional[Union[reverb.Client, reverb.TFClient]] = None,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      checkpoint: bool = True,
      save_directory: str = '~/acme',
      max_gradient_norm: Optional[float] = None,
  ):
    """Initializes the learner.

    Args:
      network: the online Q network (the one being optimized)
      target_network: the target Q critic (which lags behind the online net).
      discount: discount to use for TD updates.
      importance_sampling_exponent: power to which importance weights are raised
        before normalizing.
      learning_rate: learning rate for the q-network update.
      target_update_period: number of learner steps to perform before updating
        the target networks.
      dataset: dataset to learn from, whether fixed or from a replay buffer (see
        `acme.datasets.reverb.make_dataset` documentation).
      huber_loss_parameter: Quadratic-linear boundary for Huber loss.
      replay_client: client to replay to allow for updating priorities.
      counter: Counter object for (potentially distributed) counting.
      logger: Logger object for writing logs to.
      checkpoint: boolean indicating whether to checkpoint the learner.
      save_directory: string indicating where the learner should save
        checkpoints and snapshots.
      max_gradient_norm: used for gradient clipping.
    """

    # TODO(mwhoffman): stop allowing replay_client to be passed as a TFClient.
    # This is just here for backwards compatability for agents which reuse this
    # Learner and still pass a TFClient instance.
    if isinstance(replay_client, reverb.TFClient):
      # TODO(b/170419518): open source pytype does not understand this
      # isinstance() check because it does not have a way of getting precise
      # type information for pip-installed packages.
      replay_client = reverb.Client(replay_client._server_address)  # pytype: disable=attribute-error

    # Internalise agent components (replay buffer, networks, optimizer).
    # TODO(b/155086959): Fix type stubs and remove.
    self._iterator = iter(dataset)  # pytype: disable=wrong-arg-types
    self._network = network
    self._target_network = target_network
    self._optimizer = snt.optimizers.Adam(learning_rate)
    self._replay_client = replay_client

    # Make sure to initialize the optimizer so that its variables (e.g. the Adam
    # moments) are included in the state returned by the learner (which can then
    # be checkpointed and restored).
    self._optimizer._initialize(network.trainable_variables)  # pylint: disable= protected-access

    # Internalise the hyperparameters.
    self._discount = discount
    self._target_update_period = target_update_period
    self._importance_sampling_exponent = importance_sampling_exponent
    self._huber_loss_parameter = huber_loss_parameter
    if max_gradient_norm is None:
      max_gradient_norm = 1e10  # A very large number. Infinity results in NaNs.
    self._max_gradient_norm = tf.convert_to_tensor(max_gradient_norm)

    # Learner state.
    self._variables: List[List[tf.Tensor]] = [network.trainable_variables]
    self._num_steps = tf.Variable(0, dtype=tf.int32)

    # Internalise logging/counting objects.
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.TerminalLogger('learner', time_delta=1.)

    # Create a snapshotter object.
    if checkpoint:
      self._snapshotter = tf2_savers.Snapshotter(
          objects_to_save={'network': network},
          directory=save_directory,
          time_delta_minutes=60.)
    else:
      self._snapshotter = None

    # Do not record timestamps until after the first learning step is done.
    # This is to avoid including the time it takes for actors to come online and
    # fill the replay buffer.
    self._timestamp = None

  @tf.function
  def _step(self) -> Dict[str, tf.Tensor]:
    """Do a step of SGD and update the priorities."""

    # Pull out the data needed for updates/priorities.
    inputs = next(self._iterator)
    transitions: types.Transition = inputs.data
    keys, probs = inputs.info[:2]

    with tf.GradientTape() as tape:
      # Evaluate our networks.
      q_tm1 = self._network(transitions.observation)
      q_t_value = self._target_network(transitions.next_observation)
      q_t_selector = self._network(transitions.next_observation)

      # The rewards and discounts have to have the same type as network values.
      r_t = tf.cast(transitions.reward, q_tm1.dtype)
      r_t = tf.clip_by_value(r_t, -1., 1.)
      d_t = tf.cast(transitions.discount, q_tm1.dtype) * tf.cast(
          self._discount, q_tm1.dtype)

      # Compute the loss.
      _, extra = trfl.double_qlearning(q_tm1, transitions.action, r_t, d_t,
                                       q_t_value, q_t_selector)
      loss = losses.huber(extra.td_error, self._huber_loss_parameter)

      # Get the importance weights.
      importance_weights = 1. / probs  # [B]
      importance_weights **= self._importance_sampling_exponent
      importance_weights /= tf.reduce_max(importance_weights)

      # Reweight.
      loss *= tf.cast(importance_weights, loss.dtype)  # [B]
      loss = tf.reduce_mean(loss, axis=[0])  # []

    # Do a step of SGD.
    gradients = tape.gradient(loss, self._network.trainable_variables)
    gradients, _ = tf.clip_by_global_norm(gradients, self._max_gradient_norm)
    self._optimizer.apply(gradients, self._network.trainable_variables)

    # Get the priorities that we'll use to update.
    priorities = tf.abs(extra.td_error)

    # Periodically update the target network.
    if tf.math.mod(self._num_steps, self._target_update_period) == 0:
      for src, dest in zip(self._network.variables,
                           self._target_network.variables):
        dest.assign(src)
    self._num_steps.assign_add(1)

    # Report loss & statistics for logging.
    fetches = {
        'loss': loss,
        'keys': keys,
        'priorities': priorities,
    }

    return fetches

  def step(self):
    # Do a batch of SGD.
    result = self._step()

    # Get the keys and priorities.
    keys = result.pop('keys')
    priorities = result.pop('priorities')

    # Update the priorities in the replay buffer.
    if self._replay_client:
      self._replay_client.mutate_priorities(
          table=adders.DEFAULT_PRIORITY_TABLE,
          updates=dict(zip(keys.numpy(), priorities.numpy())))

    # Compute elapsed time.
    timestamp = time.time()
    elapsed_time = timestamp - self._timestamp if self._timestamp else 0
    self._timestamp = timestamp

    # Update our counts and record it.
    counts = self._counter.increment(steps=1, walltime=elapsed_time)
    result.update(counts)

    # Snapshot and attempt to write logs.
    if self._snapshotter is not None:
      self._snapshotter.save()
    self._logger.write(result)

  def get_variables(self, names: List[str]) -> List[np.ndarray]:
    return tf2_utils.to_numpy(self._variables)

  @property
  def state(self):
    """Returns the stateful parts of the learner for checkpointing."""
    return {
        'network': self._network,
        'target_network': self._target_network,
        'optimizer': self._optimizer,
        'num_steps': self._num_steps
    }



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AIL discriminator losses."""

import functools
from typing import Callable, Dict, Optional, Tuple

from acme import types
from acme.jax import networks as networks_lib
import jax
import jax.numpy as jnp
import tensorflow_probability as tfp
import tree

tfp = tfp.experimental.substrates.jax
tfd = tfp.distributions

# The loss is a function taking the discriminator, its state, the demo
# transition and the replay buffer transitions.
# It returns the loss as a float and a debug dictionary with the new state.
State = networks_lib.Params
DiscriminatorOutput = Tuple[networks_lib.Logits, State]
DiscriminatorFn = Callable[[State, types.Transition], DiscriminatorOutput]
Metrics = Dict[str, float]
LossOutput = Tuple[float, Tuple[Metrics, State]]
Loss = Callable[[
    DiscriminatorFn, State, types.Transition, types.Transition, networks_lib
    .PRNGKey
], LossOutput]


def _binary_cross_entropy_loss(logit: jnp.ndarray,
                               label: jnp.ndarray) -> jnp.ndarray:
  return label * jax.nn.softplus(-logit) + (1 - label) * jax.nn.softplus(logit)


@jax.vmap
def _weighted_average(x: jnp.ndarray, y: jnp.ndarray,
                      lambdas: jnp.ndarray) -> jnp.ndarray:
  return lambdas * x + (1. - lambdas) * y


def _label_data(
    rb_transitions: types.Transition,
    demonstration_transitions: types.Transition, mixup_alpha: Optional[float],
    key: networks_lib.PRNGKey) -> Tuple[types.Transition, jnp.ndarray]:
  """Create a tuple data, labels by concatenating the rb and dem transitions."""
  data = tree.map_structure(lambda x, y: jnp.concatenate([x, y]),
                            rb_transitions, demonstration_transitions)
  labels = jnp.concatenate([
      jnp.zeros(rb_transitions.reward.shape),
      jnp.ones(demonstration_transitions.reward.shape)
  ])

  if mixup_alpha is not None:
    lambda_key, mixup_key = jax.random.split(key)

    lambdas = tfd.Beta(mixup_alpha, mixup_alpha).sample(
        len(labels), seed=lambda_key)

    shuffled_data = tree.map_structure(
        lambda x: jax.random.permutation(key=mixup_key, x=x), data)
    shuffled_labels = jax.random.permutation(key=mixup_key, x=labels)

    data = tree.map_structure(lambda x, y: _weighted_average(x, y, lambdas),
                              data, shuffled_data)
    labels = _weighted_average(labels, shuffled_labels, lambdas)

  return data, labels


def _logit_bernoulli_entropy(logits: networks_lib.Logits) -> jnp.ndarray:
  return (1. - jax.nn.sigmoid(logits)) * logits - jax.nn.log_sigmoid(logits)


def gail_loss(entropy_coefficient: float = 0.,
              mixup_alpha: Optional[float] = None) -> Loss:
  """Computes the standard GAIL loss."""

  def loss_fn(
      discriminator_fn: DiscriminatorFn,
      discriminator_state: State,
      demo_transitions: types.Transition, rb_transitions: types.Transition,
      rng_key: networks_lib.PRNGKey) -> LossOutput:

    data, labels = _label_data(
        rb_transitions=rb_transitions,
        demonstration_transitions=demo_transitions,
        mixup_alpha=mixup_alpha,
        key=rng_key)
    logits, discriminator_state = discriminator_fn(discriminator_state, data)

    classification_loss = jnp.mean(_binary_cross_entropy_loss(logits, labels))

    entropy = jnp.mean(_logit_bernoulli_entropy(logits))
    entropy_loss = -entropy_coefficient * entropy

    total_loss = classification_loss + entropy_loss

    metrics = {
        'total_loss': total_loss,
        'entropy_loss': entropy_loss,
        'classification_loss': classification_loss
    }
    return total_loss, (metrics, discriminator_state)

  return loss_fn


def pugail_loss(positive_class_prior: float,
                entropy_coefficient: float,
                pugail_beta: Optional[float] = None) -> Loss:
  """Computes the PUGAIL loss (https://arxiv.org/pdf/1911.00459.pdf)."""

  def loss_fn(
      discriminator_fn: DiscriminatorFn,
      discriminator_state: State,
      demo_transitions: types.Transition, rb_transitions: types.Transition,
      rng_key: networks_lib.PRNGKey) -> LossOutput:
    del rng_key

    demo_logits, discriminator_state = discriminator_fn(discriminator_state,
                                                        demo_transitions)
    rb_logits, discriminator_state = discriminator_fn(discriminator_state,
                                                      rb_transitions)

    # Quick Maths:
    # output = logit(D) = ln(D) - ln(1-D)
    # -softplus(-output) = ln(D)
    # softplus(output) = -ln(1-D)

    # prior * -ln(D(expert))
    positive_loss = positive_class_prior * -jax.nn.log_sigmoid(demo_logits)
    # -ln(1 - D(policy)) - prior * -ln(1 - D(expert))
    negative_loss = jax.nn.softplus(
        rb_logits) - positive_class_prior * jax.nn.softplus(demo_logits)
    if pugail_beta is not None:
      negative_loss = jnp.clip(negative_loss, a_min=-1. * pugail_beta)

    classification_loss = jnp.mean(positive_loss + negative_loss)

    entropy = jnp.mean(
        _logit_bernoulli_entropy(jnp.concatenate([demo_logits, rb_logits])))
    entropy_loss = -entropy_coefficient * entropy

    total_loss = classification_loss + entropy_loss

    metrics = {
        'total_loss': total_loss,
        'positive_loss': jnp.mean(positive_loss),
        'negative_loss': jnp.mean(negative_loss),
        'demo_logits': jnp.mean(demo_logits),
        'rb_logits': jnp.mean(rb_logits),
        'entropy_loss': entropy_loss,
        'classification_loss': classification_loss
    }
    return total_loss, (metrics, discriminator_state)

  return loss_fn


def _make_gradient_penalty_data(rb_transitions: types.Transition,
                                demonstration_transitions: types.Transition,
                                key: networks_lib.PRNGKey) -> types.Transition:
  lambdas = tfd.Uniform().sample(len(rb_transitions.reward), seed=key)
  return tree.map_structure(lambda x, y: _weighted_average(x, y, lambdas),
                            rb_transitions, demonstration_transitions)


@functools.partial(jax.vmap, in_axes=(0, None, None))
def _compute_gradient_penalty(gradient_penalty_data: types.Transition,
                              discriminator_fn: Callable[[types.Transition],
                                                         float],
                              gradient_penalty_target: float) -> float:
  """Computes a penalty based on the gradient norm on the data."""
  # The input should not be batched.
  assert not gradient_penalty_data.reward.shape
  discriminator_gradient_fn = jax.grad(discriminator_fn)
  gradients = discriminator_gradient_fn(gradient_penalty_data)
  gradients = tree.map_structure(lambda x: x.flatten(), gradients)
  gradients = jnp.concatenate([gradients.observation, gradients.action,
                               gradients.next_observation])
  gradient_norms = jnp.linalg.norm(gradients + 1e-8)
  k = gradient_penalty_target * jnp.ones_like(gradient_norms)
  return jnp.mean(jnp.square(gradient_norms - k))


def add_gradient_penalty(base_loss: Loss,
                         gradient_penalty_coefficient: float,
                         gradient_penalty_target: float) -> Loss:
  """Adds a gradient penalty to the base_loss."""

  if not gradient_penalty_coefficient:
    return base_loss

  def loss_fn(discriminator_fn: DiscriminatorFn,
              discriminator_state: State,
              demo_transitions: types.Transition,
              rb_transitions: types.Transition,
              rng_key: networks_lib.PRNGKey) -> LossOutput:
    super_key, gradient_penalty_key = jax.random.split(rng_key)

    partial_loss, (losses, discriminator_state) = base_loss(
        discriminator_fn, discriminator_state, demo_transitions, rb_transitions,
        super_key)

    gradient_penalty_data = _make_gradient_penalty_data(
        rb_transitions=rb_transitions,
        demonstration_transitions=demo_transitions,
        key=gradient_penalty_key)
    def apply_discriminator_fn(transitions: types.Transition) -> float:
      logits, _ = discriminator_fn(discriminator_state, transitions)
      return logits
    gradient_penalty = gradient_penalty_coefficient * jnp.mean(
        _compute_gradient_penalty(gradient_penalty_data, apply_discriminator_fn,
                                  gradient_penalty_target))

    losses['gradient_penalty'] = gradient_penalty
    total_loss = partial_loss + gradient_penalty
    losses['total_loss'] = total_loss

    return total_loss, (losses, discriminator_state)

  return loss_fn



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""JAX networks implemented with Haiku."""

from acme.jax.networks.atari import AtariTorso
from acme.jax.networks.atari import DeepIMPALAAtariNetwork
from acme.jax.networks.atari import dqn_atari_network
from acme.jax.networks.atari import R2D2AtariNetwork
from acme.jax.networks.base import Action
from acme.jax.networks.base import FeedForwardNetwork
from acme.jax.networks.base import Logits
from acme.jax.networks.base import LogProbFn
from acme.jax.networks.base import LSTMOutputs
from acme.jax.networks.base import NetworkOutput
from acme.jax.networks.base import Observation
from acme.jax.networks.base import Params
from acme.jax.networks.base import PolicyValueRNN
from acme.jax.networks.base import PRNGKey
from acme.jax.networks.base import QNetwork
from acme.jax.networks.base import RecurrentQNetwork
from acme.jax.networks.base import SampleFn
from acme.jax.networks.base import Value
from acme.jax.networks.continuous import LayerNormMLP
from acme.jax.networks.continuous import NearZeroInitializedLinear
from acme.jax.networks.distributional import CategoricalHead
from acme.jax.networks.distributional import CategoricalValueHead
from acme.jax.networks.distributional import DiscreteValued
from acme.jax.networks.distributional import GaussianMixture
from acme.jax.networks.distributional import MultivariateNormalDiagHead
from acme.jax.networks.distributional import NormalTanhDistribution
from acme.jax.networks.duelling import DuellingMLP
from acme.jax.networks.multiplexers import CriticMultiplexer
from acme.jax.networks.policy_value import PolicyValueHead
from acme.jax.networks.rescaling import ClipToSpec
from acme.jax.networks.rescaling import TanhToSpec



# Lint as: python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility classes for saving model checkpoints."""

import datetime
import os
import pickle
from typing import Any

from absl import logging
from acme import core
from acme.tf import savers as tf_savers
import jax.numpy as jnp
import numpy as np
import tree

# Internal imports.

CheckpointState = Any

_DEFAULT_CHECKPOINT_TTL = int(datetime.timedelta(days=5).total_seconds())
_ARRAY_NAME = 'array_nest'
_EXEMPLAR_NAME = 'nest_exemplar'


def restore_from_path(ckpt_dir: str) -> CheckpointState:
  """Restore the state stored in ckpt_dir."""
  array_path = os.path.join(ckpt_dir, _ARRAY_NAME)
  exemplar_path = os.path.join(ckpt_dir, _EXEMPLAR_NAME)

  with open(exemplar_path, 'rb') as f:
    exemplar = pickle.load(f)

  with open(array_path, 'rb') as f:
    files = np.load(f, allow_pickle=True)
    flat_state = [files[key] for key in files.files]
  unflattened_tree = tree.unflatten_as(exemplar, flat_state)

  def maybe_convert_to_python(value, numpy):
    return value if numpy else value.item()

  return tree.map_structure(maybe_convert_to_python, unflattened_tree, exemplar)


def save_to_path(ckpt_dir: str, state: CheckpointState):
  """Save the state in ckpt_dir."""

  if not os.path.exists(ckpt_dir):
    os.makedirs(ckpt_dir)

  is_numpy = lambda x: isinstance(x, (np.ndarray, jnp.DeviceArray))
  flat_state = tree.flatten(state)
  nest_exemplar = tree.map_structure(is_numpy, state)

  array_path = os.path.join(ckpt_dir, _ARRAY_NAME)
  logging.info('Saving flattened array nest to %s', array_path)
  def _disabled_seek(*_):
    raise AttributeError('seek() is disabled on this object.')
  with open(array_path, 'wb') as f:
    setattr(f, 'seek', _disabled_seek)
    np.savez(f, *flat_state)

  exemplar_path = os.path.join(ckpt_dir, _EXEMPLAR_NAME)
  logging.info('Saving nest exemplar to %s', exemplar_path)
  with open(exemplar_path, 'wb') as f:
    pickle.dump(nest_exemplar, f)


# Use TF checkpointer.
class Checkpointer(tf_savers.Checkpointer):

  def __init__(
      self,
      object_to_save: core.Saveable,
      directory: str = '~/acme',
      subdirectory: str = 'default',
      **tf_checkpointer_kwargs):
    super().__init__(dict(saveable=object_to_save),
                     directory=directory,
                     subdirectory=subdirectory,
                     **tf_checkpointer_kwargs)


CheckpointingRunner = tf_savers.CheckpointingRunner



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Some types/assumptions used in the IMPALA agent."""
from typing import Callable, Tuple

from acme.agents.jax.actor_core import RecurrentState
from acme.jax import networks
import jax.numpy as jnp


# Only simple observations & discrete action spaces for now.
Observation = jnp.ndarray
Action = int
Outputs = Tuple[Tuple[networks.Logits, networks.Value], RecurrentState]
PolicyValueInitFn = Callable[[networks.PRNGKey, RecurrentState],
                             networks.Params]
PolicyValueFn = Callable[[networks.Params, Observation, RecurrentState],
                         Outputs]
RecurrentStateInitFn = Callable[[networks.PRNGKey], networks.Params]
RecurrentStateFn = Callable[[networks.Params], RecurrentState]



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Program definition for a distributed layout based on a builder."""

import logging
from typing import Any, Callable, Optional, Sequence

from acme import core
from acme import environment_loop
from acme import specs
from acme.agents.jax import builders
from acme.jax import networks as networks_lib
from acme.jax import savers
from acme.jax import types
from acme.jax import utils
from acme.utils import counting
from acme.utils import loggers
from acme.utils import lp_utils
import dm_env
import jax
import launchpad as lp
import reverb


ActorId = int
AgentNetwork = Any
NetworkFactory = Callable[[specs.EnvironmentSpec], AgentNetwork]
PolicyFactory = Callable[[AgentNetwork], Any]
EnvironmentFactory = Callable[[], dm_env.Environment]


def get_default_logger_fn(
    log_to_bigtable: bool = False,
    log_every: float = 10) -> Callable[[ActorId], loggers.Logger]:
  """Creates an actor logger."""

  def create_logger(actor_id: ActorId):
    return loggers.make_default_logger(
        'actor',
        save_data=(log_to_bigtable and actor_id == 0),
        time_delta=log_every,
        steps_key='actor_steps')
  return create_logger


def default_evaluator(environment_factory: EnvironmentFactory,
                      network_factory: NetworkFactory,
                      builder: builders.GenericActorLearnerBuilder,
                      policy_factory: PolicyFactory,
                      log_to_bigtable: bool = False) -> types.EvaluatorFactory:
  """Returns a default evaluator process."""
  def evaluator(
      random_key: networks_lib.PRNGKey,
      variable_source: core.VariableSource,
      counter: counting.Counter,
  ):
    """The evaluation process."""

    # Create environment and evaluator networks
    environment = environment_factory()
    networks = network_factory(specs.make_environment_spec(environment))

    actor = builder.make_actor(
        random_key, policy_factory(networks), variable_source=variable_source)

    # Create logger and counter.
    counter = counting.Counter(counter, 'evaluator')
    logger = loggers.make_default_logger('evaluator', log_to_bigtable,
                                         steps_key='actor_steps')

    # Create the run loop and return it.
    return environment_loop.EnvironmentLoop(environment, actor, counter,
                                              logger)
  return evaluator


class DistributedLayout:
  """Program definition for a distributed agent based on a builder."""

  def __init__(
      self,
      seed: int,
      environment_factory: EnvironmentFactory,
      network_factory: NetworkFactory,
      builder: builders.GenericActorLearnerBuilder,
      policy_network: PolicyFactory,
      num_actors: int,
      environment_spec: Optional[specs.EnvironmentSpec] = None,
      actor_logger_fn: Callable[[ActorId],
                                loggers.Logger] = get_default_logger_fn(),
      evaluator_factories: Sequence[types.EvaluatorFactory] = (),
      device_prefetch: bool = True,
      prefetch_size: int = 1,
      log_to_bigtable: bool = False,
      max_number_of_steps: Optional[int] = None,
      workdir: str = '~/acme',
      multithreading_colocate_learner_and_reverb: bool = False):
    if prefetch_size < 0:
      raise ValueError(f'Prefetch size={prefetch_size} should be non negative')

    self._seed = seed
    self._builder = builder
    self._environment_factory = environment_factory
    self._network_factory = network_factory
    self._policy_network = policy_network
    self._environment_spec = environment_spec
    self._num_actors = num_actors
    self._device_prefetch = device_prefetch
    self._log_to_bigtable = log_to_bigtable
    self._prefetch_size = prefetch_size
    self._max_number_of_steps = max_number_of_steps
    self._workdir = workdir
    self._actor_logger_fn = actor_logger_fn
    self._evaluator_factories = evaluator_factories
    self._multithreading_colocate_learner_and_reverb = (
        multithreading_colocate_learner_and_reverb)

  def replay(self):
    """The replay storage."""
    environment_spec = (
        self._environment_spec or
        specs.make_environment_spec(self._environment_factory()))
    return self._builder.make_replay_tables(environment_spec)

  def counter(self):
    kwargs = {'directory': self._workdir, 'add_uid': self._workdir == '~/acme'}
    return savers.CheckpointingRunner(
        counting.Counter(),
        key='counter',
        subdirectory='counter',
        time_delta_minutes=5,
        **kwargs)

  def learner(
      self,
      random_key: networks_lib.PRNGKey,
      replay: reverb.Client,
      counter: counting.Counter,
  ):
    """The Learning part of the agent."""

    iterator = self._builder.make_dataset_iterator(replay)

    environment_spec = (
        self._environment_spec or
        specs.make_environment_spec(self._environment_factory()))

    # Creates the networks to optimize (online) and target networks.
    networks = self._network_factory(environment_spec)

    if self._prefetch_size > 1:
      # When working with single GPU we should prefetch to device for
      # efficiency. If running on TPU this isn't necessary as the computation
      # and input placement can be done automatically. For multi-gpu currently
      # the best solution is to pre-fetch to host although this may change in
      # the future.
      device = jax.devices()[0] if self._device_prefetch else None
      iterator = utils.prefetch(
          iterator, buffer_size=self._prefetch_size, device=device)
    else:
      logging.info('Not prefetching the iterator.')

    counter = counting.Counter(counter, 'learner')

    learner = self._builder.make_learner(random_key, networks, iterator, replay,
                                         counter)
    kwargs = {'directory': self._workdir, 'add_uid': self._workdir == '~/acme'}
    # Return the learning agent.
    return savers.CheckpointingRunner(
        learner,
        key='learner',
        subdirectory='learner',
        time_delta_minutes=5,
        **kwargs)

  def actor(self, random_key: networks_lib.PRNGKey, replay: reverb.Client,
            variable_source: core.VariableSource, counter: counting.Counter,
            actor_id: int) -> environment_loop.EnvironmentLoop:
    """The actor process."""
    adder = self._builder.make_adder(replay)

    # Create environment and policy core.
    environment = self._environment_factory()

    networks = self._network_factory(specs.make_environment_spec(environment))
    policy_network = self._policy_network(networks)
    actor = self._builder.make_actor(random_key, policy_network, adder,
                                     variable_source)

    # Create logger and counter.
    counter = counting.Counter(counter, 'actor')
    # Only actor #0 will write to bigtable in order not to spam it too much.
    logger = self._actor_logger_fn(actor_id)
    # Create the loop to connect environment and agent.
    return environment_loop.EnvironmentLoop(environment, actor, counter,
                                              logger)

  def coordinator(self, counter: counting.Counter, max_actor_steps: int):
    return lp_utils.StepsLimiter(counter, max_actor_steps)

  def build(self, name='agent', program: Optional[lp.Program] = None):
    """Build the distributed agent topology."""
    if not program:
      program = lp.Program(name=name)

    key = jax.random.PRNGKey(self._seed)

    replay_node = lp.ReverbNode(self.replay)
    with program.group('replay'):
      if self._multithreading_colocate_learner_and_reverb:
        replay = replay_node.create_handle()
      else:
        replay = program.add_node(replay_node)

    with program.group('counter'):
      counter = program.add_node(lp.CourierNode(self.counter))
      if self._max_number_of_steps is not None:
        _ = program.add_node(
            lp.CourierNode(self.coordinator, counter,
                           self._max_number_of_steps))

    learner_key, key = jax.random.split(key)
    learner_node = lp.CourierNode(self.learner, learner_key, replay, counter)
    with program.group('learner'):
      if self._multithreading_colocate_learner_and_reverb:
        learner = learner_node.create_handle()
        program.add_node(
            lp.MultiThreadingColocation([learner_node, replay_node]))
      else:
        learner = program.add_node(learner_node)

    with program.group('evaluator'):
      for evaluator in self._evaluator_factories:
        evaluator_key, key = jax.random.split(key)
        program.add_node(
            lp.CourierNode(evaluator, evaluator_key, learner, counter))

    with program.group('actor'):
      for actor_id in range(self._num_actors):
        actor_key, key = jax.random.split(key)
        program.add_node(
            lp.CourierNode(self.actor, actor_key, replay, learner, counter,
                           actor_id))

    return program



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A simple agent-environment training loop."""

import operator
import time
from typing import Optional

from acme import core
from acme.utils import counting
from acme.utils import loggers
from acme.utils import signals

import dm_env
from dm_env import specs
import numpy as np
import tree


class EnvironmentLoop(core.Worker):
  """A simple RL environment loop.

  This takes `Environment` and `Actor` instances and coordinates their
  interaction. Agent is updated if `should_update=True`. This can be used as:

    loop = EnvironmentLoop(environment, actor)
    loop.run(num_episodes)

  A `Counter` instance can optionally be given in order to maintain counts
  between different Acme components. If not given a local Counter will be
  created to maintain counts between calls to the `run` method.

  A `Logger` instance can also be passed in order to control the output of the
  loop. If not given a platform-specific default logger will be used as defined
  by utils.loggers.make_default_logger. A string `label` can be passed to easily
  change the label associated with the default logger; this is ignored if a
  `Logger` instance is given.
  """

  def __init__(
      self,
      environment: dm_env.Environment,
      actor: core.Actor,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      should_update: bool = True,
      label: str = 'environment_loop',
  ):
    # Internalize agent and environment.
    self._environment = environment
    self._actor = actor
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger(label)
    self._should_update = should_update

  def run_episode(self) -> loggers.LoggingData:
    """Run one episode.

    Each episode is a loop which interacts first with the environment to get an
    observation and then give that observation to the agent in order to retrieve
    an action.

    Returns:
      An instance of `loggers.LoggingData`.
    """
    # Reset any counts and start the environment.
    start_time = time.time()
    episode_steps = 0

    # For evaluation, this keeps track of the total undiscounted reward
    # accumulated during the episode.
    episode_return = tree.map_structure(_generate_zeros_from_spec,
                                        self._environment.reward_spec())
    timestep = self._environment.reset()

    # Make the first observation.
    self._actor.observe_first(timestep)

    # Run an episode.
    while not timestep.last():
      # Generate an action from the agent's policy and step the environment.
      action = self._actor.select_action(timestep.observation)
      timestep = self._environment.step(action)

      # Have the agent observe the timestep and let the actor update itself.
      self._actor.observe(action, next_timestep=timestep)
      if self._should_update:
        self._actor.update()

      # Book-keeping.
      episode_steps += 1

      # Equivalent to: episode_return += timestep.reward
      # We capture the return value because if timestep.reward is a JAX
      # DeviceArray, episode_return will not be mutated in-place. (In all other
      # cases, the returned episode_return will be the same object as the
      # argument episode_return.)
      episode_return = tree.map_structure(operator.iadd,
                                          episode_return,
                                          timestep.reward)

    # Record counts.
    counts = self._counter.increment(episodes=1, steps=episode_steps)

    # Collect the results and combine with counts.
    steps_per_second = episode_steps / (time.time() - start_time)
    result = {
        'episode_length': episode_steps,
        'episode_return': episode_return,
        'steps_per_second': steps_per_second,
    }
    result.update(counts)
    return result

  def run(self,
          num_episodes: Optional[int] = None,
          num_steps: Optional[int] = None):
    """Perform the run loop.

    Run the environment loop either for `num_episodes` episodes or for at
    least `num_steps` steps (the last episode is always run until completion,
    so the total number of steps may be slightly more than `num_steps`).
    At least one of these two arguments has to be None.

    Upon termination of an episode a new episode will be started. If the number
    of episodes and the number of steps are not given then this will interact
    with the environment infinitely.

    Args:
      num_episodes: number of episodes to run the loop for.
      num_steps: minimal number of steps to run the loop for.

    Raises:
      ValueError: If both 'num_episodes' and 'num_steps' are not None.
    """

    if not (num_episodes is None or num_steps is None):
      raise ValueError('Either "num_episodes" or "num_steps" should be None.')

    def should_terminate(episode_count: int, step_count: int) -> bool:
      return ((num_episodes is not None and episode_count >= num_episodes) or
              (num_steps is not None and step_count >= num_steps))

    episode_count, step_count = 0, 0
    with signals.runtime_terminator():
      while not should_terminate(episode_count, step_count):
        result = self.run_episode()
        episode_count += 1
        step_count += result['episode_length']
        # Log the given episode results.
        self._logger.write(result)


def _generate_zeros_from_spec(spec: specs.Array) -> np.ndarray:
  return np.zeros(spec.shape, spec.dtype)




# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for actors."""
from typing import Optional, Tuple

from absl.testing import absltest
from absl.testing import parameterized
from acme import environment_loop
from acme import specs
from acme.agents.jax import actor_core as actor_core_lib
from acme.agents.jax import actors
from acme.jax import utils
from acme.jax import variable_utils
from acme.testing import fakes
import dm_env
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np


def _make_fake_env() -> dm_env.Environment:
  env_spec = specs.EnvironmentSpec(
      observations=specs.Array(shape=(10, 5), dtype=np.float32),
      actions=specs.DiscreteArray(num_values=3),
      rewards=specs.Array(shape=(), dtype=np.float32),
      discounts=specs.BoundedArray(
          shape=(), dtype=np.float32, minimum=0., maximum=1.),
  )
  return fakes.Environment(env_spec, episode_length=10)


class ActorTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('policy', False),
      ('policy_with_extras', True))
  def test_feedforward(self, has_extras):
    environment = _make_fake_env()
    env_spec = specs.make_environment_spec(environment)

    def policy(inputs: jnp.ndarray):
      action_values = hk.Sequential([
          hk.Flatten(),
          hk.Linear(env_spec.actions.num_values),
      ])(
          inputs)
      action = jnp.argmax(action_values, axis=-1)
      if has_extras:
        return action, (action_values,)
      else:
        return action

    policy = hk.transform(policy)

    rng = hk.PRNGSequence(1)
    dummy_obs = utils.add_batch_dim(utils.zeros_like(env_spec.observations))
    params = policy.init(next(rng), dummy_obs)

    variable_source = fakes.VariableSource(params)
    variable_client = variable_utils.VariableClient(variable_source, 'policy')

    if has_extras:
      actor_core = actor_core_lib.batched_feed_forward_with_extras_to_actor_core(
          policy.apply)
    else:
      actor_core = actor_core_lib.batched_feed_forward_to_actor_core(
          policy.apply)
    actor = actors.GenericActor(
        actor_core,
        random_key=jax.random.PRNGKey(1),
        variable_client=variable_client)

    loop = environment_loop.EnvironmentLoop(environment, actor)
    loop.run(20)


def _transform_without_rng(f):
  return hk.without_apply_rng(hk.transform(f))


class RecurrentActorTest(absltest.TestCase):

  def test_recurrent(self):
    environment = _make_fake_env()
    env_spec = specs.make_environment_spec(environment)
    output_size = env_spec.actions.num_values
    obs = utils.add_batch_dim(utils.zeros_like(env_spec.observations))
    rng = hk.PRNGSequence(1)

    @_transform_without_rng
    def network(inputs: jnp.ndarray, state: hk.LSTMState):
      return hk.DeepRNN([hk.Reshape([-1], preserve_dims=1),
                         hk.LSTM(output_size)])(inputs, state)

    @_transform_without_rng
    def initial_state(batch_size: Optional[int] = None):
      network = hk.DeepRNN([hk.Reshape([-1], preserve_dims=1),
                            hk.LSTM(output_size)])
      return network.initial_state(batch_size)

    initial_state = initial_state.apply(initial_state.init(next(rng)), 1)
    params = network.init(next(rng), obs, initial_state)

    def policy(
        params: jnp.ndarray,
        key: jnp.ndarray,
        observation: jnp.ndarray,
        core_state: hk.LSTMState
    ) -> Tuple[jnp.ndarray, hk.LSTMState]:
      del key  # Unused for test-case deterministic policy.
      action_values, core_state = network.apply(params, observation, core_state)
      actions = jnp.argmax(action_values, axis=-1)
      return actions, core_state

    variable_source = fakes.VariableSource(params)
    variable_client = variable_utils.VariableClient(variable_source, 'policy')

    actor_core = actor_core_lib.batched_recurrent_to_actor_core(
        policy, initial_state)
    actor = actors.GenericActor(actor_core, jax.random.PRNGKey(1),
                                variable_client)

    loop = environment_loop.EnvironmentLoop(environment, actor)
    loop.run(20)


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ValueDice agent implementation, using JAX."""

from typing import Callable, Iterator, List, Optional

from acme import adders
from acme import core
from acme import specs
from acme import types
from acme.adders import reverb as adders_reverb
from acme.agents.jax import actor_core as actor_core_lib
from acme.agents.jax import actors
from acme.agents.jax import builders
from acme.agents.jax.value_dice import config as value_dice_config
from acme.agents.jax.value_dice import learning
from acme.agents.jax.value_dice import networks as value_dice_networks
from acme.datasets import reverb as datasets
from acme.jax import networks as networks_lib
from acme.jax import variable_utils
from acme.utils import counting
from acme.utils import loggers
import optax
import reverb
from reverb import rate_limiters


class ValueDiceBuilder(builders.ActorLearnerBuilder):
  """ValueDice Builder.

  This builder is an entry point for online version of ValueDice.
  For offline please use the ValueDiceLearner directly.
  """

  def __init__(
      self,
      config: value_dice_config.ValueDiceConfig,
      make_demonstrations: Callable[[int], Iterator[types.Transition]],
      logger_fn: Callable[[], loggers.Logger] = lambda: None,):
    self._make_demonstrations = make_demonstrations
    self._logger_fn = logger_fn
    self._config = config

  def make_learner(
      self,
      random_key: networks_lib.PRNGKey,
      networks: value_dice_networks.ValueDiceNetworks,
      dataset: Iterator[reverb.ReplaySample],
      replay_client: Optional[reverb.Client] = None,
      counter: Optional[counting.Counter] = None,
  ) -> core.Learner:
    iterator_demonstration = self._make_demonstrations(
        self._config.batch_size * self._config.num_sgd_steps_per_step)
    policy_optimizer = optax.adam(
        learning_rate=self._config.policy_learning_rate)
    nu_optimizer = optax.adam(learning_rate=self._config.nu_learning_rate)
    return learning.ValueDiceLearner(
        networks=networks,
        policy_optimizer=policy_optimizer,
        nu_optimizer=nu_optimizer,
        discount=self._config.discount,
        rng=random_key,
        alpha=self._config.alpha,
        policy_reg_scale=self._config.policy_reg_scale,
        nu_reg_scale=self._config.nu_reg_scale,
        num_sgd_steps_per_step=self._config.num_sgd_steps_per_step,
        iterator_replay=dataset,
        iterator_demonstrations=iterator_demonstration,
        logger=self._logger_fn(),
        counter=counter,
    )

  def make_replay_tables(
      self, environment_spec: specs.EnvironmentSpec) -> List[reverb.Table]:
    samples_per_insert_tolerance = (
        self._config.samples_per_insert_tolerance_rate *
        self._config.samples_per_insert)
    error_buffer = self._config.min_replay_size * samples_per_insert_tolerance
    limiter = rate_limiters.SampleToInsertRatio(
        min_size_to_sample=self._config.min_replay_size,
        samples_per_insert=self._config.samples_per_insert,
        error_buffer=error_buffer)
    return [reverb.Table(
        name=self._config.replay_table_name,
        sampler=reverb.selectors.Uniform(),
        remover=reverb.selectors.Fifo(),
        max_size=self._config.max_replay_size,
        rate_limiter=limiter,
        signature=adders_reverb.NStepTransitionAdder.signature(
            environment_spec))]

  def make_dataset_iterator(
      self, replay_client: reverb.Client) -> Iterator[reverb.ReplaySample]:
    """Creates a dataset iterator to use for learning."""
    dataset = datasets.make_reverb_dataset(
        table=self._config.replay_table_name,
        server_address=replay_client.server_address,
        batch_size=(
            self._config.batch_size * self._config.num_sgd_steps_per_step),
        prefetch_size=self._config.prefetch_size)
    return dataset.as_numpy_iterator()

  def make_adder(self, replay_client: reverb.Client) -> adders.Adder:
    return adders_reverb.NStepTransitionAdder(
        priority_fns={self._config.replay_table_name: None},
        client=replay_client,
        n_step=1,
        discount=self._config.discount)

  def make_actor(
      self,
      random_key: networks_lib.PRNGKey,
      policy_network,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None,
  ) -> core.Actor:
    assert variable_source is not None
    actor_core = actor_core_lib.batched_feed_forward_to_actor_core(
        policy_network)
    # Inference happens on CPU, so it's better to move variables there too.
    variable_client = variable_utils.VariableClient(variable_source, 'policy',
                                                    device='cpu')
    return actors.GenericActor(
        actor_core, random_key, variable_client, adder, backend='cpu')



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the SVG0 agent class."""

import copy
from typing import Callable, Dict, Optional

import acme
from acme import specs
from acme.agents.tf.svg0_prior import agent
from acme.tf import savers as tf2_savers
from acme.utils import counting
from acme.utils import loggers
from acme.utils import lp_utils
import dm_env
import launchpad as lp
import reverb
import sonnet as snt


class DistributedSVG0:
  """Program definition for SVG0."""

  def __init__(
      self,
      environment_factory: Callable[[bool], dm_env.Environment],
      network_factory: Callable[[specs.BoundedArray], Dict[str, snt.Module]],
      num_actors: int = 1,
      num_caches: int = 0,
      environment_spec: Optional[specs.EnvironmentSpec] = None,
      batch_size: int = 256,
      prefetch_size: int = 4,
      min_replay_size: int = 1000,
      max_replay_size: int = 1000000,
      samples_per_insert: Optional[float] = 32.0,
      sequence_length: int = 10,
      sigma: float = 0.3,
      discount: float = 0.99,
      policy_optimizer: Optional[snt.Optimizer] = None,
      critic_optimizer: Optional[snt.Optimizer] = None,
      prior_optimizer: Optional[snt.Optimizer] = None,
      distillation_cost: Optional[float] = 1e-3,
      entropy_regularizer_cost: Optional[float] = 1e-3,
      target_update_period: int = 100,
      max_actor_steps: Optional[int] = None,
      log_every: float = 10.0,
  ):

    if not environment_spec:
      environment_spec = specs.make_environment_spec(environment_factory(False))

    # TODO(mwhoffman): Make network_factory directly return the struct.
    # TODO(mwhoffman): Make the factory take the entire spec.
    def wrapped_network_factory(action_spec):
      networks_dict = network_factory(action_spec)
      networks = agent.SVG0Networks(
          policy_network=networks_dict.get('policy'),
          critic_network=networks_dict.get('critic'),
          prior_network=networks_dict.get('prior', None),)
      return networks

    self._environment_factory = environment_factory
    self._network_factory = wrapped_network_factory
    self._environment_spec = environment_spec
    self._sigma = sigma
    self._num_actors = num_actors
    self._num_caches = num_caches
    self._max_actor_steps = max_actor_steps
    self._log_every = log_every
    self._sequence_length = sequence_length

    self._builder = agent.SVG0Builder(
        # TODO(mwhoffman): pass the config dataclass in directly.
        # TODO(mwhoffman): use the limiter rather than the workaround below.
        agent.SVG0Config(
            discount=discount,
            batch_size=batch_size,
            prefetch_size=prefetch_size,
            target_update_period=target_update_period,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            prior_optimizer=prior_optimizer,
            min_replay_size=min_replay_size,
            max_replay_size=max_replay_size,
            samples_per_insert=samples_per_insert,
            sequence_length=sequence_length,
            sigma=sigma,
            distillation_cost=distillation_cost,
            entropy_regularizer_cost=entropy_regularizer_cost,
        ))

  def replay(self):
    """The replay storage."""
    return self._builder.make_replay_tables(self._environment_spec,
                                            self._sequence_length)

  def counter(self):
    return tf2_savers.CheckpointingRunner(counting.Counter(),
                                          time_delta_minutes=1,
                                          subdirectory='counter')

  def coordinator(self, counter: counting.Counter):
    return lp_utils.StepsLimiter(counter, self._max_actor_steps)

  def learner(
      self,
      replay: reverb.Client,
      counter: counting.Counter,
  ):
    """The Learning part of the agent."""

    # Create the networks to optimize (online) and target networks.
    online_networks = self._network_factory(self._environment_spec.actions)
    target_networks = copy.deepcopy(online_networks)

    # Initialize the networks.
    online_networks.init(self._environment_spec)
    target_networks.init(self._environment_spec)

    dataset = self._builder.make_dataset_iterator(replay)
    counter = counting.Counter(counter, 'learner')
    logger = loggers.make_default_logger(
        'learner', time_delta=self._log_every, steps_key='learner_steps')

    return self._builder.make_learner(
        networks=(online_networks, target_networks),
        dataset=dataset,
        counter=counter,
        logger=logger,
    )

  def actor(
      self,
      replay: reverb.Client,
      variable_source: acme.VariableSource,
      counter: counting.Counter,
  ) -> acme.EnvironmentLoop:
    """The actor process."""

    # Create the behavior policy.
    networks = self._network_factory(self._environment_spec.actions)
    networks.init(self._environment_spec)
    policy_network = networks.make_policy()

    # Create the agent.
    actor = self._builder.make_actor(
        policy_network=policy_network,
        adder=self._builder.make_adder(replay),
        variable_source=variable_source,
    )

    # Create the environment.
    environment = self._environment_factory(False)

    # Create logger and counter; actors will not spam bigtable.
    counter = counting.Counter(counter, 'actor')
    logger = loggers.make_default_logger(
        'actor',
        save_data=False,
        time_delta=self._log_every,
        steps_key='actor_steps')

    # Create the loop to connect environment and agent.
    return acme.EnvironmentLoop(environment, actor, counter, logger)

  def evaluator(
      self,
      variable_source: acme.VariableSource,
      counter: counting.Counter,
      logger: Optional[loggers.Logger] = None,
  ):
    """The evaluation process."""

    # Create the behavior policy.
    networks = self._network_factory(self._environment_spec.actions)
    networks.init(self._environment_spec)
    policy_network = networks.make_policy()

    # Create the agent.
    actor = self._builder.make_actor(
        policy_network=policy_network,
        variable_source=variable_source,
        deterministic_policy=True,
    )

    # Make the environment.
    environment = self._environment_factory(True)

    # Create logger and counter.
    counter = counting.Counter(counter, 'evaluator')
    logger = logger or loggers.make_default_logger(
        'evaluator',
        time_delta=self._log_every,
        steps_key='evaluator_steps',
    )

    # Create the run loop and return it.
    return acme.EnvironmentLoop(environment, actor, counter, logger)

  def build(self, name='svg0'):
    """Build the distributed agent topology."""
    program = lp.Program(name=name)

    with program.group('replay'):
      replay = program.add_node(lp.ReverbNode(self.replay))

    with program.group('counter'):
      counter = program.add_node(lp.CourierNode(self.counter))

    if self._max_actor_steps:
      with program.group('coordinator'):
        _ = program.add_node(lp.CourierNode(self.coordinator, counter))

    with program.group('learner'):
      learner = program.add_node(lp.CourierNode(self.learner, replay, counter))

    with program.group('evaluator'):
      program.add_node(lp.CourierNode(self.evaluator, learner, counter))

    if not self._num_caches:
      # Use our learner as a single variable source.
      sources = [learner]
    else:
      with program.group('cacher'):
        # Create a set of learner caches.
        sources = []
        for _ in range(self._num_caches):
          cacher = program.add_node(
              lp.CacherNode(
                  learner, refresh_interval_ms=2000, stale_after_ms=4000))
          sources.append(cacher)

    with program.group('actor'):
      # Add actors which pull round-robin from our variable sources.
      for actor_id in range(self._num_actors):
        source = sources[actor_id % len(sources)]
        program.add_node(lp.CourierNode(self.actor, replay, source, counter))

    return program



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ValueDice config."""

import dataclasses

from acme.adders import reverb as adders_reverb


@dataclasses.dataclass
class ValueDiceConfig:
  """Configuration options for ValueDice."""

  policy_learning_rate: float = 1e-5
  nu_learning_rate: float = 1e-3
  discount: float = .99
  batch_size: int = 256
  alpha: float = 0.05
  policy_reg_scale: float = 1e-4
  nu_reg_scale: float = 10.0

  # Replay options
  replay_table_name: str = adders_reverb.DEFAULT_PRIORITY_TABLE
  samples_per_insert: float = 256 * 4
  # Rate to be used for the SampleToInsertRatio rate limitter tolerance.
  # See a formula in make_replay_tables for more details.
  samples_per_insert_tolerance_rate: float = 0.1
  min_replay_size: int = 1000
  max_replay_size: int = 1000000
  prefetch_size: int = 4

  # How many gradient updates to perform per step.
  num_sgd_steps_per_step: int = 1



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Integration test for the distributed agent."""

import functools
from absl.testing import absltest
import acme
from acme.agents.jax import r2d2
from acme.testing import fakes
import launchpad as lp


class DistributedAgentTest(absltest.TestCase):
  """Simple integration/smoke test for the distributed agent."""

  def test_agent(self):

    env_factory = lambda x: fakes.fake_atari_wrapped(oar_wrapper=True)

    config = r2d2.R2D2Config(
        batch_size=1,
        trace_length=5,
        sequence_period=1,
        samples_per_insert=1.,
        min_replay_size=32,
        burn_in_length=1,
        prefetch_size=2,
        target_update_period=2500,
        max_replay_size=100_000,
        importance_sampling_exponent=0.6,
        priority_exponent=0.9,
        max_priority_weight=0.9,
        bootstrap_n=5,
        clip_rewards=False,
        variable_update_period=400)

    agent = r2d2.DistributedR2D2FromConfig(
        environment_factory=env_factory,
        environment_spec=acme.make_environment_spec(env_factory(False)),
        network_factory=functools.partial(r2d2.make_atari_networks,
                                          config.batch_size),
        config=config,
        seed=0,
        num_actors=1,
    )

    program = agent.build()
    (learner_node,) = program.groups['learner']
    learner_node.disable_run()  # pytype: disable=attribute-error

    lp.launch(program, launch_type='test_mt')

    learner: acme.Learner = learner_node.create_handle().dereference()

    for _ in range(5):
      learner.step()


if __name__ == '__main__':
  absltest.main()



# Lint as: python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to compute running statistics."""

import dataclasses
from typing import Any, Optional, Tuple, Union

from acme import types
from acme.jax import utils
from acme.utils import tree_utils
import chex
import jax
import jax.numpy as jnp
import numpy as np
import tree


Path = Tuple[Any, ...]
"""Path in a nested structure.

  A path is a tuple of indices (normally strings for maps and integers for
  arrays and tuples) that uniquely identifies a subtree in the nested structure.
  See
  https://tree.readthedocs.io/en/latest/api.html#tree.map_structure_with_path
  for more details.
"""


def _is_prefix(a: Path, b: Path) -> bool:
  """Returns whether `a` is a prefix of `b`."""
  return b[:len(a)] == a


@chex.dataclass(frozen=True)
class NestedMeanStd:
  """A container for running statistics (mean, std) of possibly nested data."""
  mean: types.NestedArray
  std: types.NestedArray


@chex.dataclass(frozen=True)
class RunningStatisticsState(NestedMeanStd):
  """Full state of running statistics computation."""
  count: Union[int, jnp.ndarray]
  summed_variance: types.NestedArray


@dataclasses.dataclass(frozen=True)
class NestStatisticsConfig:
  """Specifies how to compute statistics for Nests with the same structure.

  Attributes:
    paths: A sequence of Nest paths to compute statistics for. If there is a
      collision between paths (one is a prefix of the other), the shorter path
      takes precedence.
  """
  paths: Tuple[Path, ...] = ((),)


def _is_path_included(config: NestStatisticsConfig, path: Path) -> bool:
  """Returns whether the path is included in the config."""
  # A path is included in the config if it corresponds to a tree node that
  # belongs to a subtree rooted at the node corresponding to some path in
  # the config.
  return any(_is_prefix(config_path, path) for config_path in config.paths)


def init_state(nest: types.Nest) -> RunningStatisticsState:
  """Initializes the running statistics for the given nested structure."""
  dtype = jnp.float64 if jax.config.jax_enable_x64 else jnp.float32

  return RunningStatisticsState(
      count=0.,
      mean=utils.zeros_like(nest, dtype=dtype),
      summed_variance=utils.zeros_like(nest, dtype=dtype),
      # Initialize with ones to make sure normalization works correctly
      # in the initial state.
      std=utils.ones_like(nest, dtype=dtype))


def _validate_batch_shapes(batch: types.NestedArray,
                           reference_sample: types.NestedArray,
                           batch_dims: Tuple[int, ...]) -> None:
  """Verifies shapes of the batch leaves against the reference sample.

  Checks that batch dimensions are the same in all leaves in the batch.
  Checks that non-batch dimensions for all leaves in the batch are the same
  as in the reference sample.

  Arguments:
    batch: the nested batch of data to be verified.
    reference_sample: the nested array to check non-batch dimensions.
    batch_dims: a Tuple of indices of batch dimensions in the batch shape.

  Returns:
    None.
  """
  def validate_node_shape(reference_sample: jnp.ndarray,
                          batch: jnp.ndarray) -> None:
    expected_shape = batch_dims + reference_sample.shape
    assert batch.shape == expected_shape, f'{batch.shape} != {expected_shape}'

  tree_utils.fast_map_structure(validate_node_shape, reference_sample, batch)


def update(state: RunningStatisticsState,
           batch: types.NestedArray,
           *,
           config: NestStatisticsConfig = NestStatisticsConfig(),
           weights: Optional[jnp.ndarray] = None,
           std_min_value: float = 1e-6,
           std_max_value: float = 1e6,
           pmap_axis_name: Optional[str] = None,
           validate_shapes: bool = True) -> RunningStatisticsState:
  """Updates the running statistics with the given batch of data.

  Note: data batch and state elements (mean, etc.) must have the same structure.

  Note: by default will use int32 for counts and float32 for accumulated
  variance. This results in an integer overflow after 2^31 data points and
  degrading precision after 2^24 batch updates or even earlier if variance
  updates have large dynamic range.
  To improve precision, consider setting jax_enable_x64 to True, see
  https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html#double-64bit-precision

  Arguments:
    state: The running statistics before the update.
    batch: The data to be used to update the running statistics.
    config: The config that specifies which leaves of the nested structure
      should the running statistics be computed for.
    weights: Weights of the batch data. Should match the batch dimensions.
      Passing a weight of 2. should be equivalent to updating on the
      corresponding data point twice.
    std_min_value: Minimum value for the standard deviation.
    std_max_value: Maximum value for the standard deviation.
    pmap_axis_name: Name of the pmapped axis, if any.
    validate_shapes: If true, the shapes of all leaves of the batch will be
      validated. Enabled by default. Doesn't impact performance when jitted.

  Returns:
    Updated running statistics.
  """
  # We require exactly the same structure to avoid issues when flattened
  # batch and state have different order of elements.
  tree.assert_same_structure(batch, state.mean)
  batch_shape = tree.flatten(batch)[0].shape
  # We assume the batch dimensions always go first.
  batch_dims = batch_shape[:len(batch_shape) - tree.flatten(state.mean)[0].ndim]
  batch_axis = range(len(batch_dims))
  if weights is None:
    step_increment = np.prod(batch_dims)
  else:
    step_increment = jnp.sum(weights)
  if pmap_axis_name is not None:
    step_increment = jax.lax.psum(step_increment, axis_name=pmap_axis_name)
  count = state.count + step_increment

  # Validation is important. If the shapes don't match exactly, but are
  # compatible, arrays will be silently broadcasted resulting in incorrect
  # statistics.
  if validate_shapes:
    if weights is not None:
      if weights.shape != batch_dims:
        raise ValueError(f'{weights.shape} != {batch_dims}')
    _validate_batch_shapes(batch, state.mean, batch_dims)

  def _compute_node_statistics(
      path: Path, mean: jnp.ndarray, summed_variance: jnp.ndarray,
      batch: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    assert isinstance(mean, jnp.ndarray), type(mean)
    assert isinstance(summed_variance, jnp.ndarray), type(summed_variance)
    if not _is_path_included(config, path):
      # Return unchanged.
      return mean, summed_variance
    # The mean and the sum of past variances are updated with Welford's
    # algorithm using batches (see https://stackoverflow.com/q/56402955).
    diff_to_old_mean = batch - mean
    if weights is not None:
      expanded_weights = jnp.reshape(
          weights,
          list(weights.shape) + [1] * (batch.ndim - weights.ndim))
      diff_to_old_mean = diff_to_old_mean * expanded_weights
    mean_update = jnp.sum(diff_to_old_mean, axis=batch_axis) / count
    if pmap_axis_name is not None:
      mean_update = jax.lax.psum(
          mean_update, axis_name=pmap_axis_name)
    mean = mean + mean_update

    diff_to_new_mean = batch - mean
    variance_update = diff_to_old_mean * diff_to_new_mean
    variance_update = jnp.sum(variance_update, axis=batch_axis)
    if pmap_axis_name is not None:
      variance_update = jax.lax.psum(variance_update, axis_name=pmap_axis_name)
    summed_variance = summed_variance + variance_update
    return mean, summed_variance

  updated_stats = tree_utils.fast_map_structure_with_path(
      _compute_node_statistics, state.mean, state.summed_variance, batch)
  # map_structure_up_to is slow, so shortcut if we know the input is not
  # structured.
  if isinstance(state.mean, jnp.ndarray):
    mean, summed_variance = updated_stats
  else:
    # Reshape the updated stats from `nest(mean, summed_variance)` to
    # `nest(mean), nest(summed_variance)`.
    mean, summed_variance = [
        tree.map_structure_up_to(
            state.mean, lambda s, i=idx: s[i], updated_stats)
        for idx in range(2)
    ]

  def compute_std(path: Path, summed_variance: jnp.ndarray,
                  std: jnp.ndarray) -> jnp.ndarray:
    assert isinstance(summed_variance, jnp.ndarray)
    if not _is_path_included(config, path):
      return std
    # Summed variance can get negative due to rounding errors.
    summed_variance = jnp.maximum(summed_variance, 0)
    std = jnp.sqrt(summed_variance / count)
    std = jnp.clip(std, std_min_value, std_max_value)
    return std

  std = tree_utils.fast_map_structure_with_path(compute_std, summed_variance,
                                                state.std)

  return RunningStatisticsState(
      count=count, mean=mean, summed_variance=summed_variance, std=std)


def normalize(batch: types.NestedArray,
              mean_std: NestedMeanStd,
              max_abs_value: Optional[float] = None) -> types.NestedArray:
  """Normalizes data using running statistics."""

  def normalize_leaf(data: jnp.ndarray, mean: jnp.ndarray,
                     std: jnp.ndarray) -> jnp.ndarray:
    # Only normalize inexact types.
    if not jnp.issubdtype(data.dtype, jnp.inexact):
      return data
    data = (data - mean) / std
    if max_abs_value is not None:
      # TODO(b/124318564): remove pylint directive
      data = jnp.clip(data, -max_abs_value, +max_abs_value)  # pylint: disable=invalid-unary-operand-type
    return data

  return tree_utils.fast_map_structure(normalize_leaf, batch, mean_std.mean,
                                       mean_std.std)


def denormalize(batch: types.NestedArray,
                mean_std: NestedMeanStd) -> types.NestedArray:
  """Denormalizes values in a nested structure using the given mean/std.

  Only values of inexact types are denormalized.
  See https://numpy.org/doc/stable/_images/dtype-hierarchy.png for Numpy type
  hierarchy.

  Args:
    batch: a nested structure containing batch of data.
    mean_std: mean and standard deviation used for denormalization.

  Returns:
    Nested structure with denormalized values.
  """

  def denormalize_leaf(data: jnp.ndarray, mean: jnp.ndarray,
                       std: jnp.ndarray) -> jnp.ndarray:
    # Only denormalize inexact types.
    if not np.issubdtype(data.dtype, np.inexact):
      return data
    return data * std + mean

  return tree_utils.fast_map_structure(denormalize_leaf, batch, mean_std.mean,
                                       mean_std.std)


@dataclasses.dataclass(frozen=True)
class NestClippingConfig:
  """Specifies how to clip Nests with the same structure.

  Attributes:
    path_map: A map that specifies how to clip values in Nests with the same
      structure. Keys correspond to paths in the nest. Values are maximum
      absolute values to use for clipping. If there is a collision between paths
      (one path is a prefix of the other), the behavior is undefined.
  """
  path_map: Tuple[Tuple[Path, float], ...] = ()


def get_clip_config_for_path(config: NestClippingConfig,
                             path: Path) -> NestClippingConfig:
  """Returns the config for a subtree from the leaf defined by the path."""
  # Start with an empty config.
  path_map = []
  for map_path, max_abs_value in config.path_map:
    if _is_prefix(map_path, path):
      return NestClippingConfig(path_map=(((), max_abs_value),))
    if _is_prefix(path, map_path):
      path_map.append((map_path[len(path):], max_abs_value))
  return NestClippingConfig(path_map=tuple(path_map))


def clip(batch: types.NestedArray,
         clipping_config: NestClippingConfig) -> types.NestedArray:
  """Clips the batch."""

  def max_abs_value_for_path(path: Path, x: jnp.ndarray) -> Optional[float]:
    del x  # Unused, needed by interface.
    return next((max_abs_value
                 for clipping_path, max_abs_value in clipping_config.path_map
                 if _is_prefix(clipping_path, path)), None)

  max_abs_values = tree_utils.fast_map_structure_with_path(
      max_abs_value_for_path, batch)

  def clip_leaf(data: jnp.ndarray,
                max_abs_value: Optional[float]) -> jnp.ndarray:
    if max_abs_value is not None:
      # TODO(b/124318564): remove pylint directive
      data = jnp.clip(data, -max_abs_value, +max_abs_value)  # pylint: disable=invalid-unary-operand-type
    return data

  return tree_utils.fast_map_structure(clip_leaf, batch, max_abs_values)


@dataclasses.dataclass(frozen=True)
class NestNormalizationConfig:
  """Specifies how to normalize Nests with the same structure.

  Attributes:
    stats_config: A config that defines how to compute running statistics to be
      used for normalization.
    clip_config: A config that defines how to clip normalized values.
  """
  stats_config: NestStatisticsConfig = NestStatisticsConfig()
  clip_config: NestClippingConfig = NestClippingConfig()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""R2D2 Builder."""
from typing import Callable, Iterator, List, Optional

import acme
from acme import adders
from acme import core
from acme import specs
from acme.adders import reverb as adders_reverb
from acme.agents.jax import actors
from acme.agents.jax import builders
from acme.agents.jax.r2d2 import actor as r2d2_actor
from acme.agents.jax.r2d2 import config as r2d2_config
from acme.agents.jax.r2d2 import learning as r2d2_learning
from acme.agents.jax.r2d2 import networks as r2d2_networks
from acme.datasets import reverb as datasets
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.jax import variable_utils
from acme.utils import counting
from acme.utils import loggers
import jax
import optax
import reverb


class R2D2Builder(builders.ActorLearnerBuilder):
  """R2D2 Builder.

  This is constructs all of the components for Recurrent Experience Replay in
  Distributed Reinforcement Learning (Kapturowski et al.)
  https://openreview.net/pdf?id=r1lyTjAqYX.
  """

  def __init__(self,
               networks: r2d2_networks.R2D2Networks,
               config: r2d2_config.R2D2Config,
               logger_fn: Callable[[], loggers.Logger] = lambda: None,):
    """Creates a R2D2 learner, a behavior policy and an eval actor.

    Args:
      networks: R2D2 networks, used to build core state spec.
      config: a config with R2D2 hps
      logger_fn: a logger factory for the learner
    """
    self._networks = networks
    self._config = config
    self._logger_fn = logger_fn

    # Sequence length for dataset iterator.
    self._sequence_length = (
        self._config.burn_in_length + self._config.trace_length + 1)

    # Construct the core state spec.
    dummy_key = jax.random.PRNGKey(0)
    initial_state_params = networks.initial_state.init(dummy_key, 1)
    initial_state = networks.initial_state.apply(initial_state_params,
                                                 dummy_key, 1)
    core_state_spec = utils.squeeze_batch_dim(initial_state)
    self._extra_spec = {'core_state': core_state_spec}

  def make_learner(
      self,
      random_key: networks_lib.PRNGKey,
      networks: r2d2_networks.R2D2Networks,
      dataset: Iterator[reverb.ReplaySample],
      replay_client: Optional[reverb.Client] = None,
      counter: Optional[counting.Counter] = None,
  ) -> core.Learner:
    # The learner updates the parameters (and initializes them).
    return r2d2_learning.R2D2Learner(
        unroll=networks.unroll,
        initial_state=networks.initial_state,
        batch_size=self._config.batch_size,
        random_key=random_key,
        burn_in_length=self._config.burn_in_length,
        discount=self._config.discount,
        importance_sampling_exponent=(
            self._config.importance_sampling_exponent),
        max_priority_weight=self._config.max_priority_weight,
        target_update_period=self._config.target_update_period,
        iterator=dataset,
        optimizer=optax.adam(self._config.learning_rate),
        bootstrap_n=self._config.bootstrap_n,
        tx_pair=self._config.tx_pair,
        clip_rewards=self._config.clip_rewards,
        replay_client=replay_client,
        counter=counter,
        logger=self._logger_fn())

  def make_replay_tables(
      self,
      environment_spec: specs.EnvironmentSpec,
  ) -> List[reverb.Table]:
    """Create tables to insert data into."""
    #'''
    if self._config.samples_per_insert:
      samples_per_insert_tolerance = (
          self._config.samples_per_insert_tolerance_rate *
          self._config.samples_per_insert)
      error_buffer = self._config.min_replay_size * samples_per_insert_tolerance
      limiter = reverb.rate_limiters.SampleToInsertRatio(
          min_size_to_sample=self._config.min_replay_size,
          samples_per_insert=self._config.samples_per_insert,
          error_buffer=error_buffer)
    else:
      limiter = reverb.rate_limiters.MinSize(self._config.min_replay_size)
    return [
        reverb.Table(
            name=self._config.replay_table_name,
            sampler=reverb.selectors.Prioritized(
                self._config.priority_exponent),
            remover=reverb.selectors.Fifo(),
            max_size=self._config.max_replay_size,
            rate_limiter=limiter,
            signature=adders_reverb.SequenceAdder.signature(
                environment_spec, self._extra_spec))
    ]

  def make_dataset_iterator(
      self, replay_client: reverb.Client) -> Iterator[reverb.ReplaySample]:
    """Create a dataset iterator to use for learning/updating the agent."""
    dataset = datasets.make_reverb_dataset(
        table=self._config.replay_table_name,
        server_address=replay_client.server_address,
        batch_size=self._config.batch_size,
        prefetch_size=self._config.prefetch_size,
        num_parallel_calls=self._config.num_parallel_calls)
    return dataset.as_numpy_iterator()

  def make_adder(self,
                 replay_client: reverb.Client) -> Optional[adders.Adder]:
    """Create an adder which records data generated by the actor/environment."""
    return adders_reverb.SequenceAdder(
        client=replay_client,
        period=self._config.sequence_period,
        sequence_length=self._sequence_length,
        delta_encoded=True)

  def make_actor(
      self,
      random_key: networks_lib.PRNGKey,
      policy_network,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None) -> acme.Actor:

    # Create variable client.
    variable_client = variable_utils.VariableClient(
        variable_source,
        key='actor_variables',
        update_period=self._config.variable_update_period)

    # TODO(b/186613827) move this to
    # - the actor __init__ function - this is a good place if it is specific
    #   for R2D2.
    # - the EnvironmentLoop - this is a good place if it potentially applies
    #   for all actors.
    #
    # Make sure not to use a random policy after checkpoint restoration by
    # assigning variables before running the environment loop.
    variable_client.update_and_wait()

    initial_state_key1, initial_state_key2, random_key = jax.random.split(
        random_key, 3)
    actor_initial_state_params = self._networks.initial_state.init(
        initial_state_key1, 1)
    actor_initial_state = self._networks.initial_state.apply(
        actor_initial_state_params, initial_state_key2, 1)

    actor_core = r2d2_actor.get_actor_core(policy_network, actor_initial_state,
                                           self._config.num_epsilons)
    return actors.GenericActor(
        actor_core, random_key, variable_client, adder, backend='cpu')



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for NStepTransition adders."""

from absl.testing import absltest
from absl.testing import parameterized
from acme.adders.reverb import test_cases
from acme.adders.reverb import test_utils
from acme.adders.reverb import transition as adders


class NStepTransitionAdderTest(test_utils.AdderTestMixin,
                               parameterized.TestCase):

  @parameterized.named_parameters(*test_cases.TEST_CASES_FOR_TRANSITION_ADDER)
  def test_adder(self, n_step, additional_discount, first, steps,
                 expected_transitions):
    adder = adders.NStepTransitionAdder(self.client, n_step,
                                        additional_discount)
    super().run_test_adder(
        adder=adder,
        first=first,
        steps=steps,
        expected_items=expected_transitions,
        stack_sequence_fields=False,
        signature=adder.signature(*test_utils.get_specs(steps[0])))


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multi-objective MPO learner implementation."""

import dataclasses
import time
from typing import Callable, List, Optional, Sequence

import acme
from acme import types
from acme.tf import losses
from acme.tf import networks
from acme.tf import savers as tf2_savers
from acme.tf import utils as tf2_utils
from acme.utils import counting
from acme.utils import loggers
import numpy as np
import sonnet as snt
import tensorflow as tf
import trfl

QValueFunctionSpec = Callable[[tf.Tensor, tf.Tensor], tf.Tensor]
RewardFunctionSpec = Callable[[tf.Tensor, tf.Tensor, tf.Tensor], tf.Tensor]

_DEFAULT_EPSILON = 1e-1
_DEFAULT_EPSILON_MEAN = 1e-3
_DEFAULT_EPSILON_STDDEV = 1e-6
_DEFAULT_INIT_LOG_TEMPERATURE = 1.
_DEFAULT_INIT_LOG_ALPHA_MEAN = 1.
_DEFAULT_INIT_LOG_ALPHA_STDDEV = 10.


@dataclasses.dataclass
class QValueObjective:
  """Defines an objective by specifying its 'Q-values' directly."""

  name: str
  # This computes "Q-values" directly from the sampled actions and other Q's.
  qvalue_fn: QValueFunctionSpec


@dataclasses.dataclass
class RewardObjective:
  """Defines an objective by specifying its reward function."""

  name: str
  # This computes the reward from observations, actions, and environment task
  # reward. In the learner, a head will automatically be added to the critic
  # network, to learn Q-values for this objective.
  reward_fn: RewardFunctionSpec


class MultiObjectiveMPOLearner(acme.Learner):
  """Distributional MPO learner.

  This is the learning component of a multi-objective MPO (MO-MPO) agent. Two
  sequences of objectives must be specified. Otherwise, the inputs are identical
  to those of the MPO / DMPO learners.

  Each objective must be defined as either a RewardObjective or an
  QValueObjective. These objectives are provided by the reward_objectives and
  qvalue_objectives parameters, respectively. For each RewardObjective, a critic
  will be trained to estimate Q-values for that objective. Whereas for each
  QValueObjective, the Q-values are computed directly by its qvalue_fn.

  A RewardObjective's reward_fn takes the observation, action, and environment
  reward as input, and returns the reward for that objective. For example, if
  the environment reward is a scalar, then an objective corresponding to the =
  task would simply return the environment reward.

  A QValueObjective's qvalue_fn takes the actions and reward-based objectives'
  Q-values as input, and outputs the "Q-values" for that objective. For
  instance, in the MO-MPO paper ([Abdolmaleki, Huang et al., 2020]), the action
  norm objective in the Humanoid run task is defined by setting the qvalue_fn
  to be the l2-norm of the actions.

  Note: If there is only one objective and that is the task reward, then this
  algorithm becomes exactly the same as (D)MPO.

  (Abdolmaleki, Huang et al., 2020): https://arxiv.org/pdf/2005.07513.pdf
  """

  def __init__(
      self,
      reward_objectives: Sequence[RewardObjective],
      qvalue_objectives: Sequence[QValueObjective],
      policy_network: snt.Module,
      critic_network: snt.Module,
      target_policy_network: snt.Module,
      target_critic_network: snt.Module,
      discount: float,
      num_samples: int,
      target_policy_update_period: int,
      target_critic_update_period: int,
      dataset: tf.data.Dataset,
      observation_network: types.TensorTransformation = tf.identity,
      target_observation_network: types.TensorTransformation = tf.identity,
      policy_loss_module: Optional[losses.MultiObjectiveMPO] = None,
      policy_optimizer: Optional[snt.Optimizer] = None,
      critic_optimizer: Optional[snt.Optimizer] = None,
      dual_optimizer: Optional[snt.Optimizer] = None,
      clipping: bool = True,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      checkpoint: bool = True,
  ):

    # Store online and target networks.
    self._policy_network = policy_network
    self._critic_network = critic_network
    self._target_policy_network = target_policy_network
    self._target_critic_network = target_critic_network

    # Make sure observation networks are snt.Module's so they have variables.
    self._observation_network = tf2_utils.to_sonnet_module(observation_network)
    self._target_observation_network = tf2_utils.to_sonnet_module(
        target_observation_network)

    # General learner book-keeping and loggers.
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger('learner')

    # Other learner parameters.
    self._discount = discount
    self._num_samples = num_samples
    self._clipping = clipping

    # Necessary to track when to update target networks.
    self._num_steps = tf.Variable(0, dtype=tf.int32)
    self._target_policy_update_period = target_policy_update_period
    self._target_critic_update_period = target_critic_update_period

    # Batch dataset and create iterator.
    # TODO(b/155086959): Fix type stubs and remove.
    self._iterator = iter(dataset)  # pytype: disable=wrong-arg-types

    # Store objectives
    self._reward_objectives = reward_objectives
    self._qvalue_objectives = qvalue_objectives
    if self._qvalue_objectives is None:
      self._qvalue_objectives = []
    self._num_critic_heads = len(self._reward_objectives)  # C
    self._objective_names = (
        [x.name for x in self._reward_objectives] +
        [x.name for x in self._qvalue_objectives])

    self._policy_loss_module = policy_loss_module or losses.MultiObjectiveMPO(
        epsilons=[losses.KLConstraint(name, _DEFAULT_EPSILON)
                  for name in self._objective_names],
        epsilon_mean=_DEFAULT_EPSILON_MEAN,
        epsilon_stddev=_DEFAULT_EPSILON_STDDEV,
        init_log_temperature=_DEFAULT_INIT_LOG_TEMPERATURE,
        init_log_alpha_mean=_DEFAULT_INIT_LOG_ALPHA_MEAN,
        init_log_alpha_stddev=_DEFAULT_INIT_LOG_ALPHA_STDDEV)

    # Check that ordering of objectives matches the policy_loss_module's
    if self._objective_names != list(self._policy_loss_module.objective_names):
      raise ValueError("Agent's ordering of objectives doesn't match "
                       "the policy loss module's ordering of epsilons.")

    # Create the optimizers.
    self._critic_optimizer = critic_optimizer or snt.optimizers.Adam(1e-4)
    self._policy_optimizer = policy_optimizer or snt.optimizers.Adam(1e-4)
    self._dual_optimizer = dual_optimizer or snt.optimizers.Adam(1e-2)

    # Expose the variables.
    policy_network_to_expose = snt.Sequential(
        [self._target_observation_network, self._target_policy_network])
    self._variables = {
        'critic': self._target_critic_network.variables,
        'policy': policy_network_to_expose.variables,
    }

    # Create a checkpointer and snapshotter object.
    self._checkpointer = None
    self._snapshotter = None

    if checkpoint:
      self._checkpointer = tf2_savers.Checkpointer(
          subdirectory='mompo_learner',
          objects_to_save={
              'counter': self._counter,
              'policy': self._policy_network,
              'critic': self._critic_network,
              'observation': self._observation_network,
              'target_policy': self._target_policy_network,
              'target_critic': self._target_critic_network,
              'target_observation': self._target_observation_network,
              'policy_optimizer': self._policy_optimizer,
              'critic_optimizer': self._critic_optimizer,
              'dual_optimizer': self._dual_optimizer,
              'policy_loss_module': self._policy_loss_module,
              'num_steps': self._num_steps,
          })

      self._snapshotter = tf2_savers.Snapshotter(
          objects_to_save={
              'policy':
                  snt.Sequential([
                      self._target_observation_network,
                      self._target_policy_network
                  ]),
          })

    # Do not record timestamps until after the first learning step is done.
    # This is to avoid including the time it takes for actors to come online and
    # fill the replay buffer.
    self._timestamp: float = None

  @tf.function
  def _step(self) -> types.NestedTensor:
    # Update target network.
    online_policy_variables = self._policy_network.variables
    target_policy_variables = self._target_policy_network.variables
    online_critic_variables = (
        *self._observation_network.variables,
        *self._critic_network.variables,
    )
    target_critic_variables = (
        *self._target_observation_network.variables,
        *self._target_critic_network.variables,
    )

    # Make online policy -> target policy network update ops.
    if tf.math.mod(self._num_steps, self._target_policy_update_period) == 0:
      for src, dest in zip(online_policy_variables, target_policy_variables):
        dest.assign(src)
    # Make online critic -> target critic network update ops.
    if tf.math.mod(self._num_steps, self._target_critic_update_period) == 0:
      for src, dest in zip(online_critic_variables, target_critic_variables):
        dest.assign(src)

    self._num_steps.assign_add(1)

    # Get data from replay (dropping extras if any). Note there is no
    # extra data here because we do not insert any into Reverb.
    inputs = next(self._iterator)
    transitions: types.Transition = inputs.data

    with tf.GradientTape(persistent=True) as tape:
      # Maybe transform the observation before feeding into policy and critic.
      # Transforming the observations this way at the start of the learning
      # step effectively means that the policy and critic share observation
      # network weights.
      o_tm1 = self._observation_network(transitions.observation)
      # This stop_gradient prevents gradients to propagate into the target
      # observation network. In addition, since the online policy network is
      # evaluated at o_t, this also means the policy loss does not influence
      # the observation network training.
      o_t = tf.stop_gradient(
          self._target_observation_network(transitions.next_observation))

      # Get online and target action distributions from policy networks.
      online_action_distribution = self._policy_network(o_t)
      target_action_distribution = self._target_policy_network(o_t)

      # Sample actions to evaluate policy; of size [N, B, ...].
      sampled_actions = target_action_distribution.sample(self._num_samples)

      # Tile embedded observations to feed into the target critic network.
      # Note: this is more efficient than tiling before the embedding layer.
      tiled_o_t = tf2_utils.tile_tensor(o_t, self._num_samples)  # [N, B, ...]

      # Compute target-estimated distributional value of sampled actions at o_t.
      sampled_q_t_all = self._target_critic_network(
          # Merge batch dimensions; to shape [N*B, ...].
          snt.merge_leading_dims(tiled_o_t, num_dims=2),
          snt.merge_leading_dims(sampled_actions, num_dims=2))

      # Compute online critic value distribution of a_tm1 in state o_tm1.
      q_tm1_all = self._critic_network(o_tm1, transitions.action)

      # Compute rewards for objectives with defined reward_fn
      reward_stats = {}
      r_t_all = []
      for objective in self._reward_objectives:
        r = objective.reward_fn(o_tm1, transitions.action, transitions.reward)
        reward_stats['{}_reward'.format(objective.name)] = tf.reduce_mean(r)
        r_t_all.append(r)
      r_t_all = tf.stack(r_t_all, axis=-1)
      r_t_all.get_shape().assert_has_rank(2)  # [B, C]

      if isinstance(sampled_q_t_all, list):  # Distributional critics
        critic_loss, sampled_q_t = _compute_distributional_critic_loss(
            sampled_q_t_all, q_tm1_all, r_t_all, transitions.discount,
            self._discount, self._num_samples)
      else:
        critic_loss, sampled_q_t = _compute_critic_loss(
            sampled_q_t_all, q_tm1_all, r_t_all, transitions.discount,
            self._discount, self._num_samples, self._num_critic_heads)

      # Add sampled Q-values for objectives with defined qvalue_fn
      sampled_q_t_k = [sampled_q_t]
      for objective in self._qvalue_objectives:
        sampled_q_t_k.append(tf.expand_dims(tf.stop_gradient(
            objective.qvalue_fn(sampled_actions, sampled_q_t)), axis=-1))
      sampled_q_t_k = tf.concat(sampled_q_t_k, axis=-1)  # [N, B, K]

      # Compute MPO policy loss.
      policy_loss, policy_stats = self._policy_loss_module(
          online_action_distribution=online_action_distribution,
          target_action_distribution=target_action_distribution,
          actions=sampled_actions,
          q_values=sampled_q_t_k)

    # For clarity, explicitly define which variables are trained by which loss.
    critic_trainable_variables = (
        # In this agent, the critic loss trains the observation network.
        self._observation_network.trainable_variables +
        self._critic_network.trainable_variables)
    policy_trainable_variables = self._policy_network.trainable_variables
    # The following are the MPO dual variables, stored in the loss module.
    dual_trainable_variables = self._policy_loss_module.trainable_variables

    # Compute gradients.
    critic_gradients = tape.gradient(critic_loss, critic_trainable_variables)
    policy_gradients, dual_gradients = tape.gradient(
        policy_loss, (policy_trainable_variables, dual_trainable_variables))

    # Delete the tape manually because of the persistent=True flag.
    del tape

    # Maybe clip gradients.
    if self._clipping:
      policy_gradients = tuple(tf.clip_by_global_norm(policy_gradients, 40.)[0])
      critic_gradients = tuple(tf.clip_by_global_norm(critic_gradients, 40.)[0])

    # Apply gradients.
    self._critic_optimizer.apply(critic_gradients, critic_trainable_variables)
    self._policy_optimizer.apply(policy_gradients, policy_trainable_variables)
    self._dual_optimizer.apply(dual_gradients, dual_trainable_variables)

    # Losses to track.
    fetches = {
        'critic_loss': critic_loss,
        'policy_loss': policy_loss,
    }
    fetches.update(policy_stats)  # Log MPO stats.
    fetches.update(reward_stats)  # Log reward stats.

    return fetches

  def step(self):
    # Run the learning step.
    fetches = self._step()

    # Compute elapsed time.
    timestamp = time.time()
    elapsed_time = timestamp - self._timestamp if self._timestamp else 0
    self._timestamp = timestamp

    # Update our counts and record it.
    counts = self._counter.increment(steps=1, walltime=elapsed_time)
    fetches.update(counts)

    # Checkpoint and attempt to write the logs.
    if self._checkpointer is not None:
      self._checkpointer.save()
    if self._snapshotter is not None:
      self._snapshotter.save()
    self._logger.write(fetches)

  def get_variables(self, names: List[str]) -> List[List[np.ndarray]]:
    return [tf2_utils.to_numpy(self._variables[name]) for name in names]


def _compute_distributional_critic_loss(
    sampled_q_t_all: List[tf.Tensor],
    q_tm1_all: List[tf.Tensor],
    r_t_all: tf.Tensor,
    d_t: tf.Tensor,
    discount: float,
    num_samples: int):
  """Compute loss and sampled Q-values for distributional critics."""
  # Compute average logits by first reshaping them and normalizing them
  # across atoms.
  batch_size = r_t_all.get_shape()[0]
  # Cast the additional discount to match the environment discount dtype.
  discount = tf.cast(discount, dtype=d_t.dtype)
  critic_losses = []
  sampled_q_ts = []
  for idx, (sampled_q_t_distributions, q_tm1_distribution) in enumerate(
      zip(sampled_q_t_all, q_tm1_all)):
    # Compute loss for distributional critic for objective c
    sampled_logits = tf.reshape(
        sampled_q_t_distributions.logits,
        [num_samples, batch_size, -1])  # [N, B, A]
    sampled_logprobs = tf.math.log_softmax(sampled_logits, axis=-1)
    averaged_logits = tf.reduce_logsumexp(sampled_logprobs, axis=0)

    # Construct the expected distributional value for bootstrapping.
    q_t_distribution = networks.DiscreteValuedDistribution(
        values=sampled_q_t_distributions.values, logits=averaged_logits)

    # Compute critic distributional loss.
    critic_loss = losses.categorical(
        q_tm1_distribution, r_t_all[:, idx], discount * d_t,
        q_t_distribution)
    critic_losses.append(tf.reduce_mean(critic_loss))

    # Compute Q-values of sampled actions and reshape to [N, B].
    sampled_q_ts.append(tf.reshape(
        sampled_q_t_distributions.mean(), (num_samples, -1)))

  critic_loss = tf.reduce_mean(critic_losses)
  sampled_q_t = tf.stack(sampled_q_ts, axis=-1)  # [N, B, C]
  return critic_loss, sampled_q_t


def _compute_critic_loss(
    sampled_q_t_all: tf.Tensor,
    q_tm1_all: tf.Tensor,
    r_t_all: tf.Tensor,
    d_t: tf.Tensor,
    discount: float,
    num_samples: int,
    num_critic_heads: int):
  """Compute loss and sampled Q-values for (non-distributional) critics."""
  # Reshape Q-value samples back to original batch dimensions and average
  # them to compute the TD-learning bootstrap target.
  batch_size = r_t_all.get_shape()[0]
  sampled_q_t = tf.reshape(
      sampled_q_t_all,
      (num_samples, batch_size, num_critic_heads))  # [N,B,C]
  q_t = tf.reduce_mean(sampled_q_t, axis=0)  # [B, C]

  # Flatten q_t and q_tm1; necessary for trfl.td_learning
  q_t = tf.reshape(q_t, [-1])  # [B*C]
  q_tm1 = tf.reshape(q_tm1_all, [-1])  # [B*C]

  # Flatten r_t_all; necessary for trfl.td_learning
  r_t_all = tf.reshape(r_t_all, [-1])  # [B*C]

  # Broadcast and then flatten d_t, to match shape of q_t and q_tm1
  d_t = tf.tile(d_t, [num_critic_heads])  # [B*C]
  # Cast the additional discount to match the environment discount dtype.
  discount = tf.cast(discount, dtype=d_t.dtype)

  # Critic loss.
  critic_loss = trfl.td_learning(q_tm1, r_t_all, discount * d_t, q_t).loss
  critic_loss = tf.reduce_mean(critic_loss)
  return critic_loss, sampled_q_t



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DQN config."""

import dataclasses
from typing import Sequence, Union

from acme.adders import reverb as adders_reverb
import jax.numpy as jnp
import numpy as np


@dataclasses.dataclass
class DQNConfig:
  """Configuration options for DQN agent.

  Attributes:
    epsilon: for use by epsilon-greedy policies. If multiple, the epsilons are
      alternated randomly per-episode.
    seed: Random seed.
    learning_rate: Learning rate for Adam optimizer.
    adam_eps: Epsilon for Adam optimizer.
    discount: Discount rate applied to value per timestep.
    n_step: N-step TD learning.
    target_update_period: Update target network every period.
    max_gradient_norm: For gradient clipping.
    batch_size: Number of transitions per batch.
    min_replay_size: Minimum replay size.
    max_replay_size: Maximum replay size.
    replay_table_name: Reverb table, defaults to DEFAULT_PRIORITY_TABLE.
    importance_sampling_exponent: Importance sampling for replay.
    priority_exponent: Priority exponent for replay.
    prefetch_size: Prefetch size for reverb replay performance.
    samples_per_insert: Ratio of learning samples to insert.
    samples_per_insert_tolerance_rate: Rate to be used for
      the SampleToInsertRatio rate limitter tolerance.
      See a formula in make_replay_tables for more details.
    num_sgd_steps_per_step: How many gradient updates to perform per learner
      step.
  """
  epsilon: Union[float, Sequence[float]] = 0.05
  # TODO(b/191706065): update all clients and remove this field.
  seed: int = 1

  # Learning rule
  learning_rate: float = 1e-3  # Learning rate for Adam optimizer.
  adam_eps: float = 1e-8  # Eps for Adam optimizer.
  discount: float = 0.99  # Discount rate applied to value per timestep.
  n_step: int = 5  # N-step TD learning.
  target_update_period: int = 100  # Update target network every period.
  max_gradient_norm: float = np.inf  # For gradient clipping.

  # Replay options
  batch_size: int = 256
  min_replay_size: int = 1_000
  max_replay_size: int = 1_000_000
  replay_table_name: str = adders_reverb.DEFAULT_PRIORITY_TABLE
  importance_sampling_exponent: float = 0.2
  priority_exponent: float = 0.6
  prefetch_size: int = 4
  samples_per_insert: float = 0.5
  samples_per_insert_tolerance_rate: float = 0.1

  num_sgd_steps_per_step: int = 1


def logspace_epsilons(num_epsilons: int, epsilon: float = 0.017
                      ) -> Sequence[float]:
  """`num_epsilons` of logspace-distributed values, with median `epsilon`."""
  if num_epsilons <= 1:
    return (epsilon,)
  return jnp.logspace(1, 8, num_epsilons, base=epsilon ** (2./9.))



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for sequence adders."""

from absl.testing import absltest
from absl.testing import parameterized

from acme.adders.reverb import sequence as adders
from acme.adders.reverb import test_cases
from acme.adders.reverb import test_utils


class SequenceAdderTest(test_utils.AdderTestMixin, parameterized.TestCase):

  @parameterized.named_parameters(*test_cases.TEST_CASES_FOR_SEQUENCE_ADDER)
  def test_adder(self,
                 sequence_length: int,
                 period: int,
                 first,
                 steps,
                 expected_sequences,
                 end_behavior: adders.EndBehavior = adders.EndBehavior.ZERO_PAD,
                 repeat_episode_times: int = 1):
    adder = adders.SequenceAdder(
        self.client,
        sequence_length=sequence_length,
        period=period,
        end_of_episode_behavior=end_behavior)
    super().run_test_adder(
        adder=adder,
        first=first,
        steps=steps,
        expected_items=expected_sequences,
        repeat_episode_times=repeat_episode_times,
        end_behavior=end_behavior,
        signature=adder.signature(*test_utils.get_specs(steps[0])))

  @parameterized.parameters(
      (True, True, adders.EndBehavior.ZERO_PAD),
      (False, True, adders.EndBehavior.TRUNCATE),
      (False, False, adders.EndBehavior.CONTINUE),
  )
  def test_end_of_episode_behavior_set_correctly(self, pad_end_of_episode,
                                                 break_end_of_episode,
                                                 expected_behavior):
    adder = adders.SequenceAdder(
        self.client,
        sequence_length=5,
        period=3,
        pad_end_of_episode=pad_end_of_episode,
        break_end_of_episode=break_end_of_episode)
    self.assertEqual(adder._end_of_episode_behavior, expected_behavior)


if __name__ == '__main__':
  absltest.main()
